import unittest

from pyansistring.helpers import (
    find_spans,
    rsearch_separators,
    search_separators,
)
from pyansistring.constants import WHITESPACE, PUNCTUATION

class HelpersTest(unittest.TestCase):
    def test_find_spans(self):
        actual = tuple(find_spans("Hello, World! Hello, World! He", "He"))
        expected = ((0, 2), (14, 16), (28, 30))
        self.assertTupleEqual(actual, expected)

    def test_search_separators(self):
        actual = tuple(
            search_separators("Hello, World!", WHITESPACE.union(PUNCTUATION))
        )
        expected = (", ", "!")
        self.assertTupleEqual(actual, expected)

    def test_rsearch_separators(self):
        actual = tuple(
            rsearch_separators("Hello, World!", WHITESPACE.union(PUNCTUATION))
        )
        expected = ("!", " ,")
        self.assertTupleEqual(actual, expected)


if __name__ == "__main__":
    unittest.main(verbosity=2)



import re
from functools import cached_property
from typing import Any, Literal

from .constants import *
from .constants import ColorMode, UnderlineMode, COLORS_8BIT, COLOR_THEMES, DEFAULT_THEME
from .frozen import FrozenMeta

class Color(metaclass=FrozenMeta):
    """Unified color representation"""
    
    def __init__(
        self,
        mode: str | None = None,
        value: Foreground | Background | Underline | tuple[int, int, int] | int | None = None
    ) -> None:
        if mode in {"4bit", "8bit", "24bit"}:
            self.mode = mode
        else:
            self.mode = None
        if isinstance(value, (Foreground, Background, Underline)):
            self.value = value.value
        elif isinstance(value, (int, tuple)):
            self.value = value
        else:
            self.value = None

    def __bool__(self) -> bool:
        return True if (self.mode and self.value) else False
    
    def __repr__(self) -> str:
        return (
            'Color('
            f'mode={self.mode!r}, '
            f'value={self.value!r})'
        )

    def __hash__(self) -> int:
        return hash((self.mode, self.value))
    
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Color):
            return NotImplemented
        return (
            (self.mode, self.value) == 
            (other.mode, other.value)
        )
    
    @classmethod
    def unset(cls) -> "Color":
        # TODO: Make it a constant?
        return cls(None, None)

    @classmethod
    def from_4bit(cls, color: Foreground | Background | Underline) -> "Color":
        return cls("4bit", color.value)
    
    @classmethod  
    def from_8bit(cls, n: int) -> "Color":
        return cls("8bit", n)
    
    @classmethod
    def from_24bit(cls, r: int, g: int, b: int) -> "Color":
        return cls("24bit", (r, g, b))
    
    def to_sgr_param(
            self,
            prefix: Literal[Foreground.SET, Background.SET, Underline.SET] | str = "",
            format_mode: Literal["standard", "compatible"] = "standard"
        ) -> str:
        if format_mode == "standard" or prefix == Underline.SET:
            separator = ":"
        else:
            separator = ";"
        if prefix:
            prefix = str(prefix) + separator
        if self.mode == "24bit" and isinstance(self.value, tuple):
            r, g, b = self.value
            return f"{prefix}2{separator*2 if separator==':' else ';'}{r}{separator}{g}{separator}{b}"
        elif self.mode == "8bit" and isinstance(self.value, int):
            return f"{prefix}5{separator}{self.value}"
        elif self.mode == "4bit" and isinstance(self.value, int):
            return f"{self.value}"
        return ""

    def to_rgb(
            self, 
            theme: Literal[
                'vga',
                'windows_xp',
                'powershell',
                'vscode',
                'windows_10',
                'terminal_app',
                'putty',
                'mirc',
                'xterm',
                'ubuntu',
                'eclipse'
            ] = DEFAULT_THEME
        ) -> tuple[int, int, int]:
        """Returns the RGB value of the color based on the theme."""
        if self.mode == "24bit":
            return self.value  # type: ignore
        elif self.mode == "8bit":
            return COLORS_8BIT[self.value]  # type: ignore
        elif self.mode == "4bit":
            return COLOR_THEMES[theme][self.value]  # type: ignore
        else:
            return (0, 0, 0)  # Default

class Style(metaclass=FrozenMeta):

    def __init__(
        self,
        foreground: Color | tuple[str, Any] = Color(),
        background: Color | tuple[str, Any] = Color(),
        underline: tuple[Color | tuple[str, Any], UnderlineMode | int] = (Color(), UnderlineMode.SINGLE),
        attributes: frozenset[SGR | int] = frozenset()
    ) -> None:
        if isinstance(foreground, tuple):
            self.foreground = Color(*foreground)
        else:
            self.foreground = foreground
        if isinstance(background, tuple):
            self.background = Color(*background)
        else:
            self.background = background
        if type(underline[1]) is int and 1 <= underline[1] <= 5:
            underline_mode = UnderlineMode(underline[1])
        elif isinstance(underline[1], UnderlineMode):
            underline_mode = underline[1]
        else:
            underline_mode = UnderlineMode.SINGLE
        if isinstance(underline[0], tuple):
            self.underline = (Color(*underline[0]), underline_mode)
        else:
            self.underline = (underline[0], underline_mode)
        self.attributes = attributes

    @cached_property
    def ansi(self) -> str:
        return self.to_ansi()

    def __bool__(self) -> bool:
        return True if (
            self.foreground 
            and self.background
            and self.underline
            and self.attributes
            ) else False
    
    def __repr__(self) -> str:
        return (
            "Style("
            f"foreground={self.foreground!r}, "
            f"background={self.background!r}, "
            f"underline=({self.underline[0]!r}, UnderlineMode.{self.underline[1].name}), "
            f"attributes={'{'}{', '.join(f'SGR.{SGR(attr).name}' for attr in self.attributes)}{'}'})"
        )
    
    def __hash__(self) -> int:
        return hash((self.foreground, self.background, self.underline, self.attributes))
    
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Style):
            return NotImplemented
        return (
            (self.foreground, self.background, self.underline, self.attributes) == 
            (other.foreground, other.background, other.underline, other.attributes)
        )
    
    def with_style(
        self,
        style: Foreground | Background | Underline | UnderlineMode | SGR | str | int | None = None,
        *args: int,
    ) -> "Style":
        """
        Returns a new Style with the given style data applied.
        """
        fg = self.foreground
        bg = self.background
        ul = self.underline
        attrs = set(self.attributes)

        if isinstance(style, Foreground):
            if style == Foreground.SET:
                if len(args) == 1:
                    fg = Color.from_8bit(args[0])
                elif len(args) == 3:
                    fg = Color.from_24bit(*args)
            else:
                fg = Color.from_4bit(style)
        elif isinstance(style, Background):
            if style == Background.SET:
                if len(args) == 1:
                    bg = Color.from_8bit(args[0])
                elif len(args) == 3:
                    bg = Color.from_24bit(*args)
            else:
                bg = Color.from_4bit(style)
        elif isinstance(style, Underline):
            if style == Underline.SET:
                if len(args) == 1:
                    ul = (Color.from_8bit(args[0]), ul[1])
                elif len(args) == 3:
                    ul = (Color.from_24bit(*args), ul[1])
            else:
                ul = (Color.from_4bit(style), ul[1])
        elif isinstance(style, UnderlineMode):
            ul = (ul[0], style)
        elif isinstance(style, SGR):
            attrs.add(style)
        elif isinstance(style, int):
            if style in SGR:
                attrs.add(SGR(style))
        elif isinstance(style, str):
            return self.from_ansi(style)

        return Style(
            foreground=fg,
            background=bg,
            underline=ul,
            attributes=frozenset(attrs)
        )
    
    def to_ansi(
        self,
        separate_codes: bool = True,
        format_mode: Literal['standard', 'compatible'] = "standard"
    ) -> str:
        parameters: list[str] = []

        if self.foreground:
            parameters.append(self.foreground.to_sgr_param(Foreground.SET, format_mode))
        if self.background:
            parameters.append(self.background.to_sgr_param(Background.SET, format_mode))
        if self.underline[0]:
            underline_mode = f"{SGR.UNDERLINE}:{self.underline[1]}"
            underline_style = f"{self.underline[0].to_sgr_param(Underline.SET, format_mode)}"
            parameters.extend((underline_mode, underline_style))

        for attr in self.attributes:
            # TODO: Should all the SGRs be at the end of the array?
            # if (attr == SGR.UNDERLINE and not self.underline[0]) \
            #     or attr in {SGR.BOLD, SGR.ITALIC}:
            #     parameters.insert(0, f"{attr}")
            # else:
                parameters.append(f"{attr}")

        if separate_codes:
            return "".join(f"\x1b[{parameter}m" for parameter in parameters)
        return f"\x1b[{';'.join(parameters)}m"

    @classmethod
    def from_ansi(cls, ansi: str) -> "Style":
        foreground = Color.unset()
        background = Color.unset()
        underline = (Color.unset(), UnderlineMode.SINGLE)
        attributes: set[SGR] = set()
        
        sequences: list[str] = re.findall(Regex.ANSI_SEQ, ansi)
        for sequence in sequences:
            sequence = (
                sequence.strip()
                    .removeprefix("\x1b[")
                    .removeprefix("\\e[")
                    .removeprefix("\033[")
                    .removesuffix("m")
                    + ";" # Add a delimiter to process the last parameter
            )
            
            parameter: str = ""
            style: Literal[Foreground.SET, Background.SET, Underline.SET, SGR.UNDERLINE] | None = None
            mode: Literal[ColorMode.PALETTE, ColorMode.TRUE_COLOR] | None = None
            rgb: list[int] = []

            for char in sequence:
                if char.isdigit():
                    parameter += char
                # Process the parameter when a delimiter is found
                elif parameter:
                    # Use a temporary variable for the check to satisfy Pylance
                    active_style = style
                    sgr_param = int(parameter)

                    if not active_style:
                        if sgr_param == Foreground.SET:
                            style = Foreground.SET
                        elif sgr_param == Background.SET:
                            style = Background.SET
                        elif sgr_param == Underline.SET:
                            style = Underline.SET
                        elif sgr_param == SGR.UNDERLINE:
                            # Set state to expect an underline mode parameter next
                            style = SGR.UNDERLINE
                        elif sgr_param in Foreground:
                            foreground = Color.from_4bit(Foreground(sgr_param))
                        elif sgr_param in Background:
                            background = Color.from_4bit(Background(sgr_param))
                        elif sgr_param in Underline:
                            underline = (Color.from_4bit(Underline(sgr_param)), underline[1])
                        elif sgr_param in SGR:
                            attributes.add(SGR(sgr_param))

                    # Check for underline mode or color mode
                    elif not mode:
                        if active_style == SGR.UNDERLINE:
                            # This special case handles codes like "4:1"
                            if 1 <= sgr_param <= 5:
                                underline = (underline[0], UnderlineMode(sgr_param))
                            else: # Fallback for simple underline
                                attributes.add(SGR.UNDERLINE)
                            style = None
                        elif sgr_param == ColorMode.PALETTE:
                            mode = ColorMode.PALETTE
                        elif sgr_param == ColorMode.TRUE_COLOR:
                            mode = ColorMode.TRUE_COLOR
                    
                    # Process color data now that style and mode are set
                    else:
                        if mode == ColorMode.PALETTE:
                            if active_style == Foreground.SET:
                                foreground = Color.from_8bit(sgr_param)
                            elif active_style == Background.SET:
                                background = Color.from_8bit(sgr_param)
                            elif active_style == Underline.SET:
                                underline = (Color.from_8bit(sgr_param), underline[1])
                            style = mode = None
                        
                        elif mode == ColorMode.TRUE_COLOR:
                            if 0 <= sgr_param <= 255:
                                rgb.append(sgr_param)
                            else:
                                # TODO: Do replace, e.g. clamp(value, 0, 255)?
                                # NOTE: Invalid RGB value, reset
                                style = mode = None
                                rgb.clear()
                                continue
                            
                            if len(rgb) == 3:
                                if active_style == Foreground.SET:
                                    foreground = Color.from_24bit(*rgb)
                                elif active_style == Background.SET:
                                    background = Color.from_24bit(*rgb)
                                elif active_style == Underline.SET:
                                    underline = (Color.from_24bit(*rgb), underline[1])
                                style = mode = None
                                rgb.clear()
                    
                    parameter = "" # Reset for the next parameter
                # If char is a delimiter but parameter is empty (e.g., "::"), do nothing.

        return cls(
            foreground=foreground,
            background=background,
            underline=underline,
            attributes=frozenset(attributes)
        )

    def merge(self, other: "Style") -> "Style":
        return Style(
            foreground=other.foreground or self.foreground,
            background=other.background or self.background,
            underline=(other.underline[0] or self.underline[0], other.underline[1]),
            attributes=other.attributes | self.attributes
        )
    
    @classmethod
    def fg_4bit(cls, color: Foreground) -> "Style":
        return cls(foreground=Color.from_4bit(color))
    
    @classmethod
    def bg_4bit(cls, color: Background) -> "Style":
        return cls(background=Color.from_4bit(color))

    @classmethod
    def ul_4bit(cls, color: Underline, mode: UnderlineMode = UnderlineMode.SINGLE) -> "Style":
        return cls(underline=(Color.from_4bit(color), mode))

    @classmethod
    def fg_8bit(cls, n: int) -> "Style":
        return cls(foreground=Color.from_8bit(n))
    
    @classmethod
    def bg_8bit(cls, n: int) -> "Style":
        return cls(background=Color.from_8bit(n))
    
    @classmethod
    def ul_8bit(cls, n: int, mode: UnderlineMode = UnderlineMode.SINGLE) -> "Style":
        return cls(underline=(Color.from_8bit(n), mode))

    @classmethod
    def fg_24bit(cls, r: int, g: int, b: int) -> "Style":
        return cls(foreground=Color.from_24bit(r, g, b))

    @classmethod
    def bg_24bit(cls, r: int, g: int, b: int) -> "Style":
        return cls(background=Color.from_24bit(r, g, b))
    
    @classmethod
    def ul_24bit(cls, r: int, g: int, b: int, mode: UnderlineMode = UnderlineMode.SINGLE) -> "Style":
        return cls(underline=(Color.from_24bit(r, g, b), mode))


if __name__ == "__main__":
    cases = (
        ("24bit", None),
        (None, 1),
        ("24bit", (10, 20, 30)),
        ("8bit", 255),
        ("4bit", Foreground.BRIGHT_GREEN),
        ("4bit", 92)
    )

    for case in cases:
        color = Color(*case)
        print(color, f"{True if color else False}")
        print(color.to_sgr_param())
        print(color.to_sgr_param(format_mode="compatible"))
        print(color.to_sgr_param(Foreground.SET))
        print(color.to_sgr_param(Background.SET))
        print(color.to_sgr_param(Underline.SET))
        print(color.to_rgb())

    print("\n")

    cases = (
        ((Foreground.BRIGHT_CYAN,),),
        ((Background.BRIGHT_YELLOW,),),
        ((Underline.SET, 255, 0, 0),),
        ((Foreground.BRIGHT_CYAN,), (Background.BRIGHT_YELLOW,)),
        ((Foreground.BRIGHT_CYAN,), (Background.BRIGHT_YELLOW,), (Underline.SET, 255, 0, 0))
    )

    style = Style()
    for case in cases:
        style = Style()
        for to_apply in case:
            style = style.with_style(*to_apply)
        print(f"\t{style.foreground = }\n\t{style.background = }\n\t{style.underline = }\n\t{style.attributes = }")
        print(f"ANSI: {repr(ansi := style.to_ansi())}")
        print(f"{ansi}Hello, World!\x1b[0m")
        print(f"FROM ANSI: {Style.from_ansi(ansi) == style}")

    merged = style.merge(Style(foreground=Color.from_8bit(165), attributes=frozenset({SGR.BOLD, SGR.ITALIC})))
    print(f"\tMERGED {merged.foreground = }\n\t{merged.background = }\n\t{merged.underline = }\n\t{merged.attributes = }")
    print(f"ANSI: {repr(ansi := merged.to_ansi())}")
    print(f"{ansi}Hello, World!\x1b[0m")



import sys
import unittest

from pyansistring import ANSIString, StyleManager # type: ignore
from pyansistring.style import Style, Color  # type: ignore
from pyansistring.constants import *


output: list[tuple[str, str, str, str]] = []


class ExtendedAssertMixin:
    def get_function_name(self, depth: int = 0) -> str:
        return sys._getframe(depth).f_code.co_name  # type: ignore

    def extended_assert_equal(
        self,
        actual: str | ANSIString,
        expected: str | ANSIString,
        verbose: bool = True,
        function_name: str | None = None,
        comment: str = "",
    ):
        if not function_name:
            function_name = self.get_function_name(depth=2)
        actual = str(actual)
        expected = str(expected)
        if verbose:
            output.append((function_name, comment, actual, expected))
        self.assertEqual(actual, expected) # type: ignore
        self.assertEqual(str(eval(repr(actual))), expected) # type: ignore


class ANSIStringFeatureTest(ExtendedAssertMixin, unittest.TestCase):
    def test_plain(self):
        actual = ANSIString("Hello, World!")
        expected = "Hello, World!"
        self.extended_assert_equal(actual, expected)

    def test_fm(self):
        bold = Style().with_style(SGR.BOLD).to_ansi()
        italic = Style().with_style(SGR.ITALIC).to_ansi()
        res = f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!").fm(SGR.BOLD),
            ANSIString("Hello, World!").fm(SGR.BOLD, (0, 5)),
            ANSIString("Hello, World!").fm(SGR.BOLD, (0, 5), slice(7, 12)),
            ANSIString("Hello, World!").fm(SGR.BOLD).fm(SGR.ITALIC),
        )
        expected = (
            "".join(f"{bold}{char}{res}" for char in "Hello, World!"),
            "".join(f"{bold}{char}{res}" for char in "Hello") + ", World!",
            "".join(f"{bold}{char}{res}" for char in "Hello")
            + ", "
            + "".join(f"{bold}{char}{res}" for char in "World")
            + "!",
            "".join(f"{bold}{italic}{char}{res}" for char in "Hello, World!"),
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_fm_w(self):
        bold = Style().with_style(SGR.BOLD).to_ansi()
        italic = Style().with_style(SGR.ITALIC).to_ansi()
        res = f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!").fm_w(SGR.BOLD, "Hello"),
            ANSIString("Hello, World!").fm_w(SGR.ITALIC, "world", case_sensitive=False),
        )
        expected = (
            "".join(f"{bold}{char}{res}" for char in "Hello") + ", World!",
            "Hello, " + "".join(f"{italic}{char}{res}" for char in "World") + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_unfm(self):
        actual = (
            ANSIString("Hello, World!").fm(SGR.BOLD).unfm(),
            ANSIString("Hello, World!").fm(SGR.BOLD, (0, 5)).unfm((0, 5)),
            ANSIString("Hello, World!")
            .fm(SGR.BOLD, (0, 5), slice(7, 12))
            .unfm((0, 5), slice(7, 12)),
        )
        expected = (
            "Hello, World!",
            "Hello, World!",
            "Hello, World!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_unfm_w(self):
        actual = (
            ANSIString("Hello, World!").fm_w(SGR.BOLD, "Hello").unfm_w("Hello"),
            ANSIString("Hello, World!")
            .fm_w(SGR.ITALIC, "world", case_sensitive=False)
            .unfm_w("world", case_sensitive=False),
        )
        expected = (
            "Hello, World!",
            "Hello, World!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_fg_4b(self):
        bright_blue = Style().with_style(Foreground.BRIGHT_BLUE).to_ansi()
        bright_yellow = Style().with_style(Foreground.BRIGHT_YELLOW).to_ansi()
        res = f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!")
            .fg_4b(Foreground.BRIGHT_BLUE, (0, 5))
            .fg_4b(Foreground.BRIGHT_YELLOW, (7, 12)),
        )
        expected = (
            "".join(f"{bright_blue}{char}{res}" for char in "Hello")
            + ", "
            + "".join(f"{bright_yellow}{char}{res}" for char in "World")
            + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_fg_4b_w(self):
        bright_blue = Style().with_style(Foreground.BRIGHT_BLUE).to_ansi()
        bright_yellow = Style().with_style(Foreground.BRIGHT_YELLOW).to_ansi()
        res = f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!")
            .fg_4b_w(Foreground.BRIGHT_BLUE, "Hello")
            .fg_4b_w(Foreground.BRIGHT_YELLOW, "World"),
        )
        expected = (
            "".join(f"{bright_blue}{char}{res}" for char in "Hello")
            + ", "
            + "".join(f"{bright_yellow}{char}{res}" for char in "World")
            + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_fg_8b(self):
        color = Style().with_style(Foreground.SET, 135).to_ansi()
        res = f"\x1b[0m"
        actual = (ANSIString("Hello, World!").fg_8b(135),)
        expected = ("".join(f"{color}{char}{res}" for char in "Hello, World!"),)
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_fg_8b_w(self):
        color = Style().with_style(Foreground.SET, 135).to_ansi()
        res = f"\x1b[0m"
        actual = (ANSIString("Hello, World!").fg_8b_w(135, "Hello, ", "World!"),)
        expected = ("".join(f"{color}{char}{res}" for char in "Hello, World!"),)
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_fg_24b(self):
        blue = Style().with_style(Foreground.SET, 0, 0, 255).to_ansi()
        yellow = Style().with_style(Foreground.SET, 255, 255, 0).to_ansi()
        res = f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!")
            .fg_24b(0, 0, 255, (0, 5))
            .fg_24b(255, 255, 0, (7, 12)),
        )
        expected = (
            "".join(f"{blue}{char}{res}" for char in "Hello")
            + ", "
            + "".join(f"{yellow}{char}{res}" for char in "World")
            + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_fg_24b_w(self):
        blue = Style().with_style(Foreground.SET, 0, 0, 255).to_ansi()
        yellow = Style().with_style(Foreground.SET, 255, 255, 0).to_ansi()
        res = f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!")
            .fg_24b_w(0, 0, 255, "Hello")
            .fg_24b_w(255, 255, 0, "World"),
        )
        expected = (
            "".join(f"{blue}{char}{res}" for char in "Hello")
            + ", "
            + "".join(f"{yellow}{char}{res}" for char in "World")
            + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_bg_4b(self):
        bright_blue = Style().with_style(Background.BRIGHT_BLUE).to_ansi()
        bright_yellow = Style().with_style(Background.BRIGHT_YELLOW).to_ansi()
        res = f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!")
            .bg_4b(Background.BRIGHT_BLUE, (0, 5))
            .bg_4b(Background.BRIGHT_YELLOW, (7, 12)),
        )
        expected = (
            "".join(f"{bright_blue}{char}{res}" for char in "Hello")
            + ", "
            + "".join(f"{bright_yellow}{char}{res}" for char in "World")
            + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_bg_4b_w(self):
        bright_blue = Style().with_style(Background.BRIGHT_BLUE).to_ansi()
        bright_yellow = Style().with_style(Background.BRIGHT_YELLOW).to_ansi()
        res = f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!")
            .bg_4b_w(Background.BRIGHT_BLUE, "Hello")
            .bg_4b_w(Background.BRIGHT_YELLOW, "World"),
        )
        expected = (
            "".join(f"{bright_blue}{char}{res}" for char in "Hello")
            + ", "
            + "".join(f"{bright_yellow}{char}{res}" for char in "World")
            + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_bg_8b(self):
        color = Style().with_style(Background.SET, 135).to_ansi()
        res = f"\x1b[0m"
        actual = (ANSIString("Hello, World!").bg_8b(135),)
        expected = ("".join(f"{color}{char}{res}" for char in "Hello, World!"),)
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_bg_8b_w(self):
        color = Style().with_style(Background.SET, 135).to_ansi()
        res = f"\x1b[0m"
        actual = (ANSIString("Hello, World!").bg_8b_w(135, "Hello, ", "World!"),)
        expected = ("".join(f"{color}{char}{res}" for char in "Hello, World!"),)
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_bg_24b(self):
        blue = Style().with_style(Background.SET, 0, 0, 255).to_ansi()
        yellow = Style().with_style(Background.SET, 255, 255, 0).to_ansi()
        res = f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!")
            .bg_24b(0, 0, 255, (0, 5))
            .bg_24b(255, 255, 0, (7, 12)),
        )
        expected = (
            "".join(f"{blue}{char}{res}" for char in "Hello")
            + ", "
            + "".join(f"{yellow}{char}{res}" for char in "World")
            + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_bg_24b_w(self):
        blue = Style().with_style(Background.SET, 0, 0, 255).to_ansi()
        yellow = Style().with_style(Background.SET, 255, 255, 0).to_ansi()
        res = f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!")
            .bg_24b_w(0, 0, 255, "Hello")
            .bg_24b_w(255, 255, 0, "World"),
        )
        expected = (
            "".join(f"{blue}{char}{res}" for char in "Hello")
            + ", "
            + "".join(f"{yellow}{char}{res}" for char in "World")
            + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_from_ansi(self):
        blue = f"\x1b[38;2;0;0;255m"
        yellow = f"\x1b[38;2;255;255;0m"
        res = f"\x1b[0m"
        plain = (
            f"\x1b[10;10H{''.join(f'{blue}{char}{res}' for char in 'Hello')}, {yellow}World{res}!{res*3}",
            ANSIString("Hello, World!")
            .fg_24b_w(0, 0, 255, "Hello")
            .fg_24b_w(255, 255, 0, "World")
            .styled_text,
        )
        styles = {index: Style.from_ansi(blue) for index in range(0, 5)} | {
            index: Style.from_ansi(yellow) for index in range(7, 12)
        }
        actual = (ANSIString.from_ansi(p) for p in plain)
        expected = ANSIString("Hello, World!", styles)
        for a in actual:
            self.extended_assert_equal(a, expected)
            self.assertDictEqual(a.style_manager, expected.style_manager)

    def test_rainbow(self):
        actual = (
            ANSIString("abcdefghijklmnopqrstuvwxyz").rainbow(skip_whitespace=True),
            ANSIString("abcdefghijklmnopqrstuvwxyz").rainbow(bg=True),
        )
        expected = (
            ANSIString(
                "abcdefghijklmnopqrstuvwxyz",
                {
                    0: "\x1b[38;2;255;0;0m", 1: "\x1b[38;2;255;60;0m",
                    2: "\x1b[38;2;255;119;0m", 3: "\x1b[38;2;255;179;0m",
                    4: "\x1b[38;2;255;234;0m", 5: "\x1b[38;2;217;255;0m",
                    6: "\x1b[38;2;157;255;0m", 7: "\x1b[38;2;98;255;0m",
                    8: "\x1b[38;2;38;255;0m", 9: "\x1b[38;2;0;255;21m",
                    10: "\x1b[38;2;0;255;76m", 11: "\x1b[38;2;0;255;136m",
                    12: "\x1b[38;2;0;255;195m", 13: "\x1b[38;2;0;255;255m",
                    14: "\x1b[38;2;0;195;255m", 15: "\x1b[38;2;0;136;255m",
                    16: "\x1b[38;2;0;76;255m", 17: "\x1b[38;2;0;21;255m",
                    18: "\x1b[38;2;38;0;255m", 19: "\x1b[38;2;98;0;255m",
                    20: "\x1b[38;2;157;0;255m", 21: "\x1b[38;2;217;0;255m",
                    22: "\x1b[38;2;255;0;234m", 23: "\x1b[38;2;255;0;178m",
                    24: "\x1b[38;2;255;0;119m", 25: "\x1b[38;2;255;0;60m",
                }
            ),
            ANSIString(
                "abcdefghijklmnopqrstuvwxyz",
                {
                    0: "\x1b[48;2;255;0;0m", 1: "\x1b[48;2;255;60;0m",
                    2: "\x1b[48;2;255;119;0m", 3: "\x1b[48;2;255;179;0m",
                    4: "\x1b[48;2;255;234;0m", 5: "\x1b[48;2;217;255;0m",
                    6: "\x1b[48;2;157;255;0m", 7: "\x1b[48;2;98;255;0m",
                    8: "\x1b[48;2;38;255;0m", 9: "\x1b[48;2;0;255;21m",
                    10: "\x1b[48;2;0;255;76m", 11: "\x1b[48;2;0;255;136m",
                    12: "\x1b[48;2;0;255;195m", 13: "\x1b[48;2;0;255;255m",
                    14: "\x1b[48;2;0;195;255m", 15: "\x1b[48;2;0;136;255m",
                    16: "\x1b[48;2;0;76;255m", 17: "\x1b[48;2;0;21;255m",
                    18: "\x1b[48;2;38;0;255m", 19: "\x1b[48;2;98;0;255m",
                    20: "\x1b[48;2;157;0;255m", 21: "\x1b[48;2;217;0;255m",
                    22: "\x1b[48;2;255;0;234m", 23: "\x1b[48;2;255;0;178m",
                    24: "\x1b[48;2;255;0;119m", 25: "\x1b[48;2;255;0;60m",
                }
            ),
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_multicolor(self):
        string = "abcdefghijklmnopqrstuvwxyz"
        actual = (
            ANSIString(string).multicolor(MulticolorSequences.RAINBOW),
            ANSIString(string).multicolor(MulticolorSequences.REVERSED_RAINBOW),
            ANSIString(string).multicolor(
                "r=84:|g=161:|b=255: $ r+9:minmax(0,inf)|g+4:minmax(0,inf) &*"
            ),
            ANSIString(string).multicolor(
                "r=84:|g=161:|b=255: $ r+9:minmax(0,inf)|g+4:minmax(0,inf) @&*"
            ),
            ANSIString(string).multicolor(
                "r=84:|g=161:|b=255: $ r+50:minmax(0,inf)|g+25:minmax(0,inf) # b-70:minmax(0,inf) !&*"
            ),
        )
        expected = (
            ANSIString(
                "abcdefghijklmnopqrstuvwxyz",
                {
                    0: "\x1b[38;2;255;0;0m", 1: "\x1b[38;2;255;51;0m",
                    2: "\x1b[38;2;255;102;0m", 3: "\x1b[38;2;255;153;0m",
                    4: "\x1b[38;2;255;204;0m", 5: "\x1b[38;2;255;255;0m",
                    6: "\x1b[38;2;191;255;0m", 7: "\x1b[38;2;127;255;0m",
                    8: "\x1b[38;2;63;255;0m", 9: "\x1b[38;2;0;255;0m",
                    10: "\x1b[38;2;0;255;63m", 11: "\x1b[38;2;0;255;127m",
                    12: "\x1b[38;2;0;255;191m", 13: "\x1b[38;2;0;255;255m",
                    14: "\x1b[38;2;0;191;255m", 15: "\x1b[38;2;0;127;255m",
                    16: "\x1b[38;2;0;63;255m", 17: "\x1b[38;2;0;0;255m",
                    18: "\x1b[38;2;63;0;255m", 19: "\x1b[38;2;127;0;255m",
                    20: "\x1b[38;2;191;0;255m", 21: "\x1b[38;2;255;0;255m",
                    22: "\x1b[38;2;255;0;191m", 23: "\x1b[38;2;255;0;127m",
                    24: "\x1b[38;2;255;0;63m", 25: "\x1b[38;2;255;0;0m",
                },
            ),
            ANSIString(
                "abcdefghijklmnopqrstuvwxyz",
                {
                    0: "\x1b[38;2;255;0;0m", 1: "\x1b[38;2;255;0;63m",
                    2: "\x1b[38;2;255;0;127m", 3: "\x1b[38;2;255;0;191m",
                    4: "\x1b[38;2;255;0;255m", 5: "\x1b[38;2;191;0;255m",
                    6: "\x1b[38;2;127;0;255m", 7: "\x1b[38;2;63;0;255m",
                    8: "\x1b[38;2;0;0;255m", 9: "\x1b[38;2;0;63;255m",
                    10: "\x1b[38;2;0;127;255m", 11: "\x1b[38;2;0;191;255m",
                    12: "\x1b[38;2;0;255;255m", 13: "\x1b[38;2;0;255;191m",
                    14: "\x1b[38;2;0;255;127m", 15: "\x1b[38;2;0;255;63m",
                    16: "\x1b[38;2;0;255;0m", 17: "\x1b[38;2;63;255;0m",
                    18: "\x1b[38;2;127;255;0m", 19: "\x1b[38;2;191;255;0m",
                    20: "\x1b[38;2;255;255;0m", 21: "\x1b[38;2;255;204;0m",
                    22: "\x1b[38;2;255;153;0m", 23: "\x1b[38;2;255;102;0m",
                    24: "\x1b[38;2;255;51;0m", 25: "\x1b[38;2;255;0;0m",
                },
            ),
            ANSIString(
                "abcdefghijklmnopqrstuvwxyz",
                {
                    0: "\x1b[38;2;84;161;255m", 1: "\x1b[38;2;93;165;255m",
                    2: "\x1b[38;2;102;169;255m", 3: "\x1b[38;2;111;173;255m",
                    4: "\x1b[38;2;120;177;255m", 5: "\x1b[38;2;129;181;255m",
                    6: "\x1b[38;2;138;185;255m", 7: "\x1b[38;2;147;189;255m",
                    8: "\x1b[38;2;156;193;255m", 9: "\x1b[38;2;165;197;255m",
                    10: "\x1b[38;2;174;201;255m", 11: "\x1b[38;2;183;205;255m",
                    12: "\x1b[38;2;192;209;255m", 13: "\x1b[38;2;201;213;255m",
                    14: "\x1b[38;2;210;217;255m", 15: "\x1b[38;2;219;221;255m",
                    16: "\x1b[38;2;228;225;255m", 17: "\x1b[38;2;237;229;255m",
                    18: "\x1b[38;2;246;233;255m", 19: "\x1b[38;2;255;237;255m",
                    20: "\x1b[38;2;255;241;255m", 21: "\x1b[38;2;255;245;255m",
                    22: "\x1b[38;2;255;249;255m", 23: "\x1b[38;2;255;253;255m",
                    24: "\x1b[38;2;255;255;255m", 25: "\x1b[38;2;255;255;255m",
                },
            ),
            ANSIString(
                "abcdefghijklmnopqrstuvwxyz",
                {
                    0: "\x1b[38;2;255;255;255m", 1: "\x1b[38;2;255;255;255m",
                    2: "\x1b[38;2;255;253;255m", 3: "\x1b[38;2;255;249;255m",
                    4: "\x1b[38;2;255;245;255m", 5: "\x1b[38;2;255;241;255m",
                    6: "\x1b[38;2;255;237;255m", 7: "\x1b[38;2;246;233;255m",
                    8: "\x1b[38;2;237;229;255m", 9: "\x1b[38;2;228;225;255m",
                    10: "\x1b[38;2;219;221;255m", 11: "\x1b[38;2;210;217;255m",
                    12: "\x1b[38;2;201;213;255m", 13: "\x1b[38;2;192;209;255m",
                    14: "\x1b[38;2;183;205;255m", 15: "\x1b[38;2;174;201;255m",
                    16: "\x1b[38;2;165;197;255m", 17: "\x1b[38;2;156;193;255m",
                    18: "\x1b[38;2;147;189;255m", 19: "\x1b[38;2;138;185;255m",
                    20: "\x1b[38;2;129;181;255m", 21: "\x1b[38;2;120;177;255m",
                    22: "\x1b[38;2;111;173;255m", 23: "\x1b[38;2;102;169;255m",
                    24: "\x1b[38;2;93;165;255m", 25: "\x1b[38;2;84;161;255m",
                },
            ),
            ANSIString(
                "abcdefghijklmnopqrstuvwxyz",
                {
                    0: "\x1b[38;2;84;161;255m", 1: "\x1b[38;2;134;186;255m",
                    2: "\x1b[38;2;134;186;185m", 3: "\x1b[38;2;134;186;255m",
                    4: "\x1b[38;2;84;161;255m", 5: "\x1b[38;2;134;186;255m",
                    6: "\x1b[38;2;134;186;185m", 7: "\x1b[38;2;134;186;255m",
                    8: "\x1b[38;2;84;161;255m", 9: "\x1b[38;2;134;186;255m",
                    10: "\x1b[38;2;134;186;185m", 11: "\x1b[38;2;134;186;255m",
                    12: "\x1b[38;2;84;161;255m", 13: "\x1b[38;2;134;186;255m",
                    14: "\x1b[38;2;134;186;185m", 15: "\x1b[38;2;134;186;255m",
                    16: "\x1b[38;2;84;161;255m", 17: "\x1b[38;2;134;186;255m",
                    18: "\x1b[38;2;134;186;185m", 19: "\x1b[38;2;134;186;255m",
                    20: "\x1b[38;2;84;161;255m", 21: "\x1b[38;2;134;186;255m",
                    22: "\x1b[38;2;134;186;185m", 23: "\x1b[38;2;134;186;255m",
                    24: "\x1b[38;2;84;161;255m", 25: "\x1b[38;2;134;186;255m",
                },
            ),
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)
        self.assertDictEqual(
            actual[0].style_manager,
            {
                key: value
                for key, value in enumerate(reversed(tuple(actual[1].style_manager.values())))
            },
        )
        self.assertDictEqual(
            actual[2].style_manager,
            {
                key: value
                for key, value in enumerate(reversed(tuple(actual[3].style_manager.values())))
            },
        )

    def test_multicolor_c(self):
        actual = (
            ANSIString("Hello, \nWorld!\n It's pyansistring!")
            .multicolor_c(MulticolorSequences.RAINBOW)
            .style_manager.values()
        )
        expected = (
            ANSIString("Hello, World! It's pyansistring!")
            .multicolor(MulticolorSequences.RAINBOW)
            .style_manager.values()
        )
        for a, e in zip(actual, expected):
            self.assertEqual(a, e)

    def test_to_svg(self):
        # Build a minimal fake TTFont replacement
        class _Glyph:
            def __init__(self, width: int) -> None:
                self.width = width

        class _Head:
            unitsPerEm = 1000

        class _Hhea:
            ascent = 800
            descent = -200
            lineGap = 200

        class _Hmtx:
            def __getitem__(self, key: str):
                return (600, 0)  # width, lsb

        class _Name:
            def getDebugName(self, _):
                return "FakeFont"

        class FakeTTFont:
            def __getitem__(self, key: str):
                if key == "head":
                    return _Head()
                if key == "hhea":
                    return _Hhea()
                if key == "hmtx":
                    return _Hmtx()
                if key == "name":
                    return _Name()
                raise KeyError(key)

            def getBestCmap(self) -> dict[int, str]:
                # Force fallback to ".notdef" for all glyphs
                return {}  # type: ignore

            def getGlyphSet(self):
                return {".notdef": _Glyph(600)}

        # Ensure to_svg does not raise ImportError
        import pyansistring.pyansistring as pas
        pas.is_fonttools_available = True

        # Case 1: fully styled string
        s1 = ANSIString("Hello").fg_24b(255, 0, 0)
        svg1 = s1.to_svg(FakeTTFont(), font_size_px=16)  # type: ignore
        # self.assertTrue(svg1.startswith('<?xml'))
        self.assertIn('<svg ', svg1)
        self.assertIn('</svg>', svg1)
        self.assertIn('font-family="FakeFont"', svg1)
        self.assertEqual(svg1.count('fill="rgb(255, 0, 0)"'), len("Hello"))

        # Case 2: partially styled string
        s2 = ANSIString("Hello, World!").fg_24b(255, 0, 0, (0, 5))
        svg2 = s2.to_svg(FakeTTFont(), font_size_px=16)  # type: ignore
        self.assertEqual(svg2.count('fill="rgb(255, 0, 0)"'), 5)
        self.assertIn("<tspan>,</tspan>", svg2)

    def test_ul_attr(self):
        ul = Style().with_style(SGR.UNDERLINE).to_ansi()
        res = "\x1b[0m"
        actual = (
            ANSIString("Hello, World!").fm(SGR.UNDERLINE),
            ANSIString("Hello, World!").fm(SGR.UNDERLINE, (0, 5)),
            ANSIString("Hello, World!").fm(SGR.UNDERLINE, (0, 5), slice(7, 12)),
        )
        expected = (
            "".join(f"{ul}{c}{res}" for c in "Hello, World!"),
            "".join(f"{ul}{c}{res}" for c in "Hello") + ", World!",
            "".join(f"{ul}{c}{res}" for c in "Hello") + ", " + "".join(f"{ul}{c}{res}" for c in "World") + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_ul_8b(self):
        ulc = Style().with_style(Underline.SET, 135).to_ansi()
        res = "\x1b[0m"
        actual = (
            ANSIString("Hello, World!").ul_8b(135),
            ANSIString("Hello, World!").ul_8b(135, (0, 5)),
            ANSIString("Hello, World!").ul_8b(135, (0, 5), slice(7, 12)),
        )
        expected = (
            "".join(f"{ulc}{c}{res}" for c in "Hello, World!"),
            "".join(f"{ulc}{c}{res}" for c in "Hello") + ", World!",
            "".join(f"{ulc}{c}{res}" for c in "Hello") + ", " + "".join(f"{ulc}{c}{res}" for c in "World") + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_ul_8b_w(self):
        ulc = Style().with_style(Underline.SET, 135).to_ansi()
        res = "\x1b[0m"
        actual = (
            ANSIString("Hello, World!").ul_8b_w(135, "Hello"),
            ANSIString("Hello, World!").ul_8b_w(135, "World"),
        )
        expected = (
            "".join(f"{ulc}{c}{res}" for c in "Hello") + ", World!",
            "Hello, " + "".join(f"{ulc}{c}{res}" for c in "World") + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_ul_24b(self):
        blue_ul = Style().with_style(Underline.SET, 0, 0, 255).to_ansi()
        yellow_ul = Style().with_style(Underline.SET, 255, 255, 0).to_ansi()
        res = "\x1b[0m"
        actual = (
            ANSIString("Hello, World!").ul_24b(0, 0, 255, (0, 5)).ul_24b(255, 255, 0, (7, 12)),
        )
        expected = (
            "".join(f"{blue_ul}{c}{res}" for c in "Hello") + ", " + "".join(f"{yellow_ul}{c}{res}" for c in "World") + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_ul_24b_w(self):
        blue_ul = Style().with_style(Underline.SET, 0, 0, 255).to_ansi()
        yellow_ul = Style().with_style(Underline.SET, 255, 255, 0).to_ansi()
        res = "\x1b[0m"
        actual = (
            ANSIString("Hello, World!").ul_24b_w(0, 0, 255, "Hello").ul_24b_w(255, 255, 0, "World"),
        )
        expected = (
            "".join(f"{blue_ul}{c}{res}" for c in "Hello") + ", " + "".join(f"{yellow_ul}{c}{res}" for c in "World") + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_ul_modes(self):
        """Test underline modes with color."""
        modes = (
            UnderlineMode.SINGLE,
            UnderlineMode.DOUBLE,
            UnderlineMode.CURLY,
            UnderlineMode.DOTTED,
            UnderlineMode.DASHED,
        )
        for mode in modes:
            with self.subTest(mode=mode.name):
                actual = ANSIString("Hello, World!").ul_24b(0, 128, 255).fm(mode)
                ul_ansi = Style().with_style(Underline.SET, 0, 128, 255).with_style(mode).to_ansi()
                expected = "".join(f"{ul_ansi}{c}\x1b[0m" for c in "Hello, World!")
                self.extended_assert_equal(actual, expected)

    def test_ul_modes_w(self):
        """Test word-targeted underline mode with color."""
        mode = UnderlineMode.DOUBLE
        actual = (
            ANSIString("Hello, World!").ul_8b_w(135, "World").fm_w(mode, "World"),
        )
        ul_ansi = Style().with_style(Underline.SET, 135).with_style(mode).to_ansi()
        expected = (
            "Hello, " + "".join(f"{ul_ansi}{c}\x1b[0m" for c in "World") + "!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test_fstring_basic(self):
        blue = Style().with_style(Foreground.BLUE).to_ansi()
        res = "\x1b[0m"
        actual = f"{ANSIString('Hello, World!').fg_4b(Foreground.BLUE)}"
        expected = "".join(f"{blue}{c}{res}" for c in "Hello, World!")
        self.extended_assert_equal(actual, expected)

    def test_fstring_with_literal_prefix(self):
        blue = Style().with_style(Foreground.BLUE).to_ansi()
        res = "\x1b[0m"
        actual = f"s{ANSIString('Hello, World!').fg_4b(Foreground.BLUE)}"
        expected = "s" + "".join(f"{blue}{c}{res}" for c in "Hello, World!")
        self.extended_assert_equal(actual, expected)

    def test_fstring_with_format_spec(self):
        red = Style().with_style(Foreground.RED).to_ansi()
        res = "\x1b[0m"
        actual = f"{ANSIString('Hi').fg_4b(Foreground.RED):>3}"
        expected = " " + "".join(f"{red}{c}{res}" for c in "Hi")
        self.extended_assert_equal(actual, expected)

if __name__ == "__main__":

    unittest.main(argv=["first-arg-is-ignored"], verbosity=2, exit=False)

    wall = "\x1b[100m \x1b[0m"
    if output:
        print(f"\n{'VERBOSE OUTPUT':-^70}")
    for function_name, comment, actual, expected in output:
        print(
            f"{function_name:<40}{'ACTUAL:':>10} "
            f"{wall}{actual}{wall}\n{comment:^35}{'EXPECTED:':>15} {wall}{expected}{wall}"
        )



__all__ = [
    "find_spans",
    "search_separators",
    "rsearch_separators",
    "clamp",
    "hsl_to_rgb",
    "ValueRange",
    "Length",
]

from collections.abc import Generator
from dataclasses import dataclass

from pyansistring.constants import WHITESPACE


def find_spans(string: str, substring: str) -> Generator[tuple[int, int], None, None]:
    i = j = 0
    while i < len(string):
        i = string[i:].find(substring)
        if i == -1:
            break
        yield (i + j, i + j + len(substring))
        j = i = i + j + len(substring)


def search_separators(string: str, allowed: set[str] = WHITESPACE):
    """Search for allowed separators in a string."""
    separator = ""
    for char in string:
        if char in allowed:
            separator += char
        elif separator:
            yield separator
            separator = ""
    if separator:
        yield separator


def rsearch_separators(string: str, allowed: set[str] = WHITESPACE):
    """Search for allowed separators in a reversed string."""
    return search_separators(string[::-1], allowed)


def clamp(
    value: int | float,
    min_: int | float = float("-inf"),
    max_: int | float = float("inf"),
) -> int | float:
    """Restrict a number between two other numbers."""
    return min_ if value < min_ else max_ if value > max_ else value


def hsl_to_rgb(
    hue: int | float, saturation: int | float = 100, lightness: int | float = 50
) -> tuple[int, int, int]:
    hue = hue / 100
    saturation = saturation / 100
    lightness = lightness / 100

    def f(n: int | float):
        k = (n + hue * (10 / 3)) % 12
        a = saturation * min(lightness, 1 - lightness)
        return round((lightness - a * max(-1, min(k - 3, 9 - k, 1))) * 255)

    return f(0), f(8), f(4)


@dataclass(frozen=True)
class ValueRange:
    lo: int
    hi: int

    def __hash__(self) -> int:
        return hash((self.lo, self.hi))


@dataclass(frozen=True)
class Length:
    value: int

    def __hash__(self) -> int:
        return hash(self.value)



"""
pyansistring - A library for string color styling using ANSI escape sequences.

This library provides a string class that inherits from Python's str and supports
ANSI styling while preserving string methods. It supports 4-, 8-, and 24-bit color
modes, per-word coloring, alignment, and automated coloring features.
"""

from importlib.metadata import PackageNotFoundError, version

__name__ = "pyansistring"

try:
    __version__ = version("pyansistring")
except PackageNotFoundError:
    __version__ = "unknown"

from .pyansistring import ANSIString
from .style import Color, Style
from .style_manager import StyleManager

from .constants import (
    ColorMode,
    Foreground,
    Background,
    Underline,
    UnderlineMode,
    SGR,
    COLORS_8BIT,
    COLOR_THEMES,
    DEFAULT_THEME,
    WHITESPACE,
    UNIVERSAL_NEWLINES,
    PUNCTUATION,
    PUNCTUATION_AND_WHITESPACE,
    Regex,
    MulticolorSequences,
)

__all__ = [
    "ANSIString",
    "Color",
    "Style",
    "StyleManager",
    "ColorMode",
    "Foreground",
    "Background",
    "Underline",
    "UnderlineMode",
    "SGR",
    "COLORS_8BIT",
    "COLOR_THEMES",
    "DEFAULT_THEME",
    "WHITESPACE",
    "UNIVERSAL_NEWLINES",
    "PUNCTUATION",
    "PUNCTUATION_AND_WHITESPACE",
    "Regex",
    "MulticolorSequences",
]



import unittest

from tests.test_pyansistring_features import ExtendedAssertMixin
from pyansistring import ANSIString, StyleManager # type: ignore
from pyansistring.style import Style, Color  # type: ignore
from pyansistring.constants import *


output: list[tuple[str, str, ANSIString, str]] = []


class ANSIStringDefaultTest(ExtendedAssertMixin, unittest.TestCase):
    def test___getitem__(self):
        bold, italic, res = f"\x1b[1m", f"\x1b[3m", f"\x1b[0m"
        actual = (
            ANSIString("Hello, World!")
            .fm(SGR.BOLD, (0, 5))
            .fm(SGR.ITALIC, (7, 12))[2:-2]
        )
        expected = (
            "".join(f"{bold}{char}{res}" for char in "llo")
            + ", "
            + "".join(f"{italic}{char}{res}" for char in "Worl")
        )
        self.extended_assert_equal(actual, expected)

    def test___add__(self):
        bold, italic, res = f"\x1b[1m", f"\x1b[3m", f"\x1b[0m"
        actual = (
            ANSIString("Hello").fm(SGR.BOLD) + ANSIString(", World!").fm(SGR.ITALIC),
            ANSIString("Hello").fm(SGR.BOLD) + ", World!",
        )
        expected = (
            "".join(f"{bold}{char}{res}" for char in "Hello")
            + "".join(f"{italic}{char}{res}" for char in ", World!"),
            "".join(f"{bold}{char}{res}" for char in "Hello") + ", World!",
        )
        for a, e in zip(actual, expected):
            self.extended_assert_equal(a, e)

    def test___radd__(self):
        bold, res = f"\x1b[1m", f"\x1b[0m"
        actual = "Hello" + ANSIString(", World!").fm(SGR.BOLD)
        expected = "Hello" + "".join(f"{bold}{char}{res}" for char in ", World!")
        self.extended_assert_equal(actual, expected)

    def test_capitalize(self):
        bold, res = f"\x1b[1m", f"\x1b[0m"
        actual = ANSIString("hello, world!").fm(SGR.BOLD).capitalize()
        expected = "".join(f"{bold}{char}{res}" for char in "Hello, world!")
        self.extended_assert_equal(actual, expected)

    def test_center(self):
        bold, res = f"\x1b[1m", f"\x1b[0m"
        actual = tuple(
            ANSIString("Hello, World!").fm(SGR.BOLD).center(width, "^")
            for width in range(13, 18)
        )
        expected = (
            "".join(f"{bold}{char}{res}" for char in "Hello, World!"),
            "".join(f"{bold}{char}{res}" for char in "Hello, World!") + "^",
            "^" + "".join(f"{bold}{char}{res}" for char in "Hello, World!") + "^",
            "^" + "".join(f"{bold}{char}{res}" for char in "Hello, World!") + "^^",
            "^^" + "".join(f"{bold}{char}{res}" for char in "Hello, World!") + "^^",
        )
        for no, (a, e) in enumerate(zip(actual, expected)):
            self.extended_assert_equal(
                a,
                e,
                verbose=(False if no != 4 else True),
                function_name=("" if no != 4 else None),
            )

    def test_endswith(self):
        string = ANSIString("Hello, World!").fm(SGR.BOLD)
        self.assertEqual(string.endswith("!"), True)
        self.assertEqual(string.endswith("[0m"), False)

    def test_find(self):
        actual = (
            ANSIString(" Hello, World!").fm(SGR.BOLD).find(" "),
            ANSIString(" Hello, World!").fm(SGR.BOLD).find(" ", 1),
        )
        expected = (0, 7)
        for a, e in zip(actual, expected):
            self.assertEqual(a, e)

    def test_index(self):
        actual = (
            ANSIString(" Hello, World!").fm(SGR.BOLD).index(" "),
            ANSIString(" Hello, World!").fm(SGR.BOLD).index(" ", 1),
        )
        expected = (0, 7)
        for a, e in zip(actual, expected):
            self.assertEqual(a, e)

    def test_is_methods(self):
        methods = (
            "isalnum",
            "isalpha",
            "isascii",
            "isdecimal",
            "isdigit",
            "isidentifier",
            "islower",
            "isnumeric",
            "isprintable",
            "isspace",
            "istitle",
            "isupper",
        )
        string = ANSIString("Hello, World!").fm(SGR.BOLD)
        actual = (getattr(string, method)() for method in methods)
        expected = (getattr(string.plain_text, method)() for method in methods)
        for a, e in zip(actual, expected):
            self.assertEqual(a, e)

    def test_join(self):
        blue, yellow = f"\x1b[38:2::0:0:255m", f"\x1b[38:2::255:255:0m"
        bold, res = f"\x1b[1m", f"\x1b[0m"
        actual = (
            ANSIString(", ")
            .fm(SGR.BOLD)
            .join(
                (
                    "Anyway",
                    ANSIString("Hello").fg_24b(0, 0, 255),
                    ANSIString("World!").fg_24b(255, 255, 0),
                )
            )
        )
        expected = (
            "Anyway"
            + "".join(f"{bold}{char}{res}" for char in ", ")
            + "".join(f"{blue}{char}{res}" for char in "Hello")
            + "".join(f"{bold}{char}{res}" for char in ", ")
            + "".join(f"{yellow}{char}{res}" for char in "World!")
        )
        self.extended_assert_equal(actual, expected)

    def test_ljust(self):
        bold, res = f"\x1b[1m", f"\x1b[0m"
        actual = tuple(
            ANSIString("Hello, World!").fm(SGR.BOLD).ljust(width, "<")
            for width in range(13, 18)
        )
        expected = (
            "".join(f"{bold}{char}{res}" for char in "Hello, World!"),
            "".join(f"{bold}{char}{res}" for char in "Hello, World!") + "<",
            "".join(f"{bold}{char}{res}" for char in "Hello, World!") + "<<",
            "".join(f"{bold}{char}{res}" for char in "Hello, World!") + "<<<",
            "".join(f"{bold}{char}{res}" for char in "Hello, World!") + "<<<<",
        )
        for no, (a, e) in enumerate(zip(actual, expected)):
            self.extended_assert_equal(
                a,
                e,
                verbose=(False if no != 4 else True),
                function_name=("" if no != 4 else None),
            )

    def test_lower(self):
        bold, res = f"\x1b[1m", f"\x1b[0m"
        actual = ANSIString("Hello, World!").fm(SGR.BOLD).lower()
        expected = "".join(f"{bold}{char}{res}" for char in "hello, world!")
        self.extended_assert_equal(actual, expected)

    def test_rfind(self):
        actual = (
            ANSIString("Hello, World! ").fm(SGR.BOLD).rfind(" "),
            ANSIString("Hello, World! ").fm(SGR.BOLD).rfind(" ", 0, 13),
        )
        expected = (13, 6)
        for a, e in zip(actual, expected):
            self.assertEqual(a, e)

    def test_rindex(self):
        actual = (
            ANSIString("Hello, World! ").fm(SGR.BOLD).rindex(" "),
            ANSIString("Hello, World! ").fm(SGR.BOLD).rindex(" ", 0, 13),
        )
        expected = (13, 6)
        for a, e in zip(actual, expected):
            self.assertEqual(a, e)

    def test_rjust(self):
        bold, res = f"\x1b[1m", f"\x1b[0m"
        actual = tuple(
            ANSIString("Hello, World!").fm(SGR.BOLD).rjust(width, ">")
            for width in range(13, 18)
        )
        expected = (
            "".join(f"{bold}{char}{res}" for char in "Hello, World!"),
            ">" + "".join(f"{bold}{char}{res}" for char in "Hello, World!"),
            ">>" + "".join(f"{bold}{char}{res}" for char in "Hello, World!"),
            ">>>" + "".join(f"{bold}{char}{res}" for char in "Hello, World!"),
            ">>>>" + "".join(f"{bold}{char}{res}" for char in "Hello, World!"),
        )
        for no, (a, e) in enumerate(zip(actual, expected)):
            self.extended_assert_equal(
                a,
                e,
                verbose=(False if no != 4 else True),
                function_name=("" if no != 4 else None),
            )

    def test_rsplit(self):
        blue, yellow, res = f"\x1b[38:2::0:0:255m", f"\x1b[38:2::255:255:0m", "\x1b[0m"
        steps = 0
        seps = (None, ".", "..", "...")
        maxsplits = (-1, 0, 1, 2, 3)
        spaces = (
            ANSIString(" hello,   world!    ")
            .fg_24b(0, 0, 255, (1, 2), (5, 6))
            .fg_24b(255, 255, 0, (10, 11), (14, 15))
        )
        dots = (
            ANSIString(".hello,...world!....")
            .fg_24b(0, 0, 255, (1, 2), (5, 6))
            .fg_24b(255, 255, 0, (10, 11), (14, 15))
        )
        hello = f"{blue}h{res}ell{blue}o{res}"
        world = f"{yellow}w{res}orl{yellow}d{res}"
        actual = {
            sep: {
                maxsplit: (spaces if not sep else dots).rsplit(sep, maxsplit)
                for maxsplit in maxsplits
            }
            for sep in seps
        }
        expected: dict[None | str, dict[int, list[str]]] = {
            None: {
                -1: [f"{hello},", f"{world}!"],
                0: [f" {hello},   {world}!"],
                1: [f" {hello},", f"{world}!"],
                2: [f"{hello},", f"{world}!"],
                3: [f"{hello},", f"{world}!"],
            },
            ".": {
                -1: ["", f"{hello},", "", "", f"{world}!", "", "", "", ""],
                0: [f".{hello},...{world}!...."],
                1: [f".{hello},...{world}!...", ""],
                2: [f".{hello},...{world}!..", "", ""],
                3: [f".{hello},...{world}!.", "", "", ""],
            },
            "..": {
                -1: [f".{hello},.", f"{world}!", "", ""],
                0: [f".{hello},...{world}!...."],
                1: [f".{hello},...{world}!..", ""],
                2: [f".{hello},...{world}!", "", ""],
                3: [f".{hello},.", f"{world}!", "", ""],
            },
            "...": {
                -1: [f".{hello},", f"{world}!.", ""],
                0: [f".{hello},...{world}!...."],
                1: [f".{hello},...{world}!.", ""],
                2: [f".{hello},", f"{world}!.", ""],
                3: [f".{hello},", f"{world}!.", ""],
            },
        }
        for sep in seps:
            for maxsplit in maxsplits:
                a_list, e_list = actual[sep][maxsplit], expected[sep][maxsplit]
                self.assertEqual(len(a_list), len(e_list))
                a_list, e_list = filter(None, a_list), filter(None, e_list)
                for a, e in zip(a_list, e_list, strict=True):
                    self.extended_assert_equal(
                        a,
                        e,
                        verbose=False,
                        comment=f"({repr(sep)} {maxsplit})",
                        function_name=(None if not steps else ""),
                    )
                    steps += 1

    def test_split(self):
        blue, yellow, res = f"\x1b[38:2::0:0:255m", f"\x1b[38:2::255:255:0m", "\x1b[0m"
        steps = 0
        seps = (None, ".", "..", "...")
        maxsplits = (-1, 0, 1, 2, 3)
        spaces = (
            ANSIString(" Hello,   World!    ")
            .fg_24b(0, 0, 255, (1, 2), (5, 6))
            .fg_24b(255, 255, 0, (10, 11), (14, 15))
        )
        dots = (
            ANSIString(".Hello,...World!....")
            .fg_24b(0, 0, 255, (1, 2), (5, 6))
            .fg_24b(255, 255, 0, (10, 11), (14, 15))
        )
        hello = f"{blue}H{res}ell{blue}o{res}"
        world = f"{yellow}W{res}orl{yellow}d{res}"
        actual = {
            sep: {
                maxsplit: (spaces if not sep else dots).split(sep, maxsplit)
                for maxsplit in maxsplits
            }
            for sep in seps
        }
        expected: dict[None | str, dict[int, list[str]]] = {
            None: {
                -1: [f"{hello},", f"{world}!"],
                0: [f"{hello},   {world}!    "],
                1: [f"{hello},", f"{world}!    "],
                2: [f"{hello},", f"{world}!"],
                3: [f"{hello},", f"{world}!"],
            },
            ".": {
                -1: ["", f"{hello},", "", "", f"{world}!", "", "", "", ""],
                0: [f".{hello},...{world}!...."],
                1: ["", f"{hello},...{world}!...."],
                2: ["", f"{hello},", f"..{world}!...."],
                3: ["", f"{hello},", "", f".{world}!...."],
            },
            "..": {
                -1: [f".{hello},", f".{world}!", "", ""],
                0: [f".{hello},...{world}!...."],
                1: [f".{hello},", f".{world}!...."],
                2: [f".{hello},", f".{world}!", ".."],
                3: [f".{hello},", f".{world}!", "", ""],
            },
            "...": {
                -1: [f".{hello},", f"{world}!", "."],
                0: [f".{hello},...{world}!...."],
                1: [f".{hello},", f"{world}!...."],
                2: [f".{hello},", f"{world}!", "."],
                3: [f".{hello},", f"{world}!", "."],
            },
        }
        for sep in seps:
            for maxsplit in maxsplits:
                a_list, e_list = actual[sep][maxsplit], expected[sep][maxsplit]
                self.assertEqual(len(a_list), len(e_list))
                a_list, e_list = filter(None, a_list), filter(None, e_list)
                for a, e in zip(a_list, e_list, strict=True):
                    self.extended_assert_equal(
                        a,
                        e,
                        verbose=False,
                        comment=f"({repr(sep)} {maxsplit})",
                        function_name=(None if not steps else ""),
                    )
                    steps += 1

    def test_splitlines(self):
        blue, yellow, res = f"\x1b[38:2::0:0:255m", f"\x1b[38:2::255:255:0m", "\x1b[0m"
        hello = f"{blue}H{res}ell{blue}o{res}"
        world = f"{yellow}W{res}orld{yellow}!{res}"
        string = (
            ANSIString("\n\nHello, \nWorld!\n\n\n")
            .fg_24b(0, 0, 255, (2, 3), (6, 7))
            .fg_24b(255, 255, 0, (10, 11), (15, 16))
        )
        actual = (
            string.splitlines(),
            string.splitlines(True),
        )
        expected = (
            ["", "", f"{hello}, ", f"{world}", "", ""],
            ["\n", "\n", f"{hello}, \n", f"{world}\n", "\n", "\n"],
        )
        for step, (a_list, e_list) in enumerate(zip(actual, expected)):
            self.assertEqual(len(a_list), len(e_list))
            a_list, e_list = filter(None, a_list), filter(None, e_list)
            for a, e in zip(a_list, e_list, strict=True):
                self.extended_assert_equal(
                    a,
                    e,
                    verbose=False,
                    comment=f"(keeptrue={False if not step else True})",
                )

    def test_startswith(self):
        string = ANSIString("Hello, World!").fm(SGR.BOLD)
        self.assertEqual(string.startswith("H"), True)
        self.assertEqual(string.startswith("\x1b"), False)

    def test_swapcase(self):
        bold, res = f"\x1b[1m", f"\x1b[0m"
        actual = ANSIString("Hello, World!").fm(SGR.BOLD).swapcase()
        expected = "".join(f"{bold}{char}{res}" for char in "hELLO, wORLD!")
        self.extended_assert_equal(actual, expected)

    def test_title(self):
        bold, res = f"\x1b[1m", f"\x1b[0m"
        actual = ANSIString("HELLO, WoRlD!").fm(SGR.BOLD).title()
        expected = "".join(f"{bold}{char}{res}" for char in "Hello, World!")
        self.extended_assert_equal(actual, expected)

    def test_upper(self):
        bold, res = f"\x1b[1m", f"\x1b[0m"
        actual = ANSIString("Hello, World!").fm(SGR.BOLD).upper()
        expected = "".join(f"{bold}{char}{res}" for char in "HELLO, WORLD!")
        self.extended_assert_equal(actual, expected)


from __future__ import annotations

__all__ = [
    "StyleManager",
    "ANSIString",
]

import re
from collections.abc import Generator, Sequence, Iterable
from copy import copy, deepcopy
from itertools import cycle
from pathlib import Path
from random import randint
from typing import Annotated, Any, Self, Union, SupportsIndex, TYPE_CHECKING

if not TYPE_CHECKING:
    try:
        from fontTools.ttLib import TTFont
        from fontTools.pens.svgPathPen import SVGPathPen
        from fontTools.pens.transformPen import TransformPen
        is_fonttools_available = True
    except Exception:
        is_fonttools_available = False
else:
    from fontTools.ttLib import TTFont
    from fontTools.pens.svgPathPen import SVGPathPen
    from fontTools.pens.transformPen import TransformPen
    is_fonttools_available = True

from .constants import *
from .helpers import *
from .style import Style
from .style_manager import StyleManager

class MulticolorInstruction:
    """Class representing a single instruction in a multicolor command."""
    color: str
    operator: str
    value: str
    processed_value: int | float
    mode: str
    minmax: tuple[float, float]
    repeat: int

    def __init__(self, rgb: dict[str, dict[str, int]], **kwargs) -> None:
        allowed_keys = {"color", "operator", "value", "mode", "minmax", "repeat"}
        missing_keys = tuple(k for k in allowed_keys if k not in kwargs)
        if not kwargs or len(missing_keys):
            raise TypeError(
                f"{self.__class__}.__init__() missing {len(missing_keys)}"
                f" required keyword argument{'s' if len(missing_keys)>1 else ''}:"
                ", ".join(missing_keys)
            )

        self.rgb = rgb
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.processed_value = self.process_value(self.value)

        if isinstance(self.minmax, str):
            self.minmax = tuple(map(float, self.minmax[7:-1].split(",")))
        elif self.minmax is None:
            self.minmax = (0, 255)

        if not self.mode:
            self.mode = "fg"

        if self.operator == ">":
            base_value = rgb[self.mode][self.color]
            if base_value <= self.processed_value:
                self.operator = "+"
                self.processed_value = (self.processed_value - base_value) / self.repeat
            elif base_value > self.processed_value:
                self.operator = "-"
                self.processed_value = (base_value - self.processed_value) / self.repeat

    def process_value(self, value: str, save: bool = False) -> int | float:
        if value.startswith("random"):
            from_value, to_value = map(int, value[7:-1].split(","))
            value = randint(from_value, to_value)
        elif value.endswith(("r", "g", "b")):
            mode, color = value.split("_")
            value = self.rgb[mode][color]
        else:
            value = float(value)
        if save:
            self.processed_value = value
        return value


class MulticolorCommand:
    """Class representing a multicolor command with multiple instructions."""
    def __init__(
        self,
        instructions: list[MulticolorInstruction] | None = None,
        reset: str | None = None,
        repeat: int | str | None = None,
    ) -> None:
        self.instructions = instructions if instructions else []
        self.reset = reset
        if isinstance(repeat, str):
            self.repeat = int(repeat[7:-1])
        else:
            self.repeat = 1 if repeat is None else repeat


class ANSIString(str):
    r"""
    String class that allows you to extend your vanilla str with ANSI escape sequences for coloring/styling.

    Instance Attributes:
        _styles: dictionary containing pairs of char indices with ANSI escape sequences.
        _styled: plain string to which ANSI e.s. from `_styles` has been applied.

    Properties:
        styles: a getter for `_styles`.
        styled: a getter for `_styled` (checks if `styles` has been modified and renders it if so).
        plain: unformatted, normal string.
        actual_length: returns the length of `styled`.

    Note:
        *The `ANSIString` class is unhashable for consistency, because `styles` is an unhashable dict
        that we can change.
    """
    _style_manager: StyleManager
    _styled_text: str

    def __new__(
        cls,
        plain_text: str = "",
        style_manager: StyleManager | dict[int, Style] | dict[int, str] | None = None
    ) -> Self:
        instance = super().__new__(cls, plain_text)
        if isinstance(style_manager, StyleManager):
            instance._style_manager = style_manager
        elif isinstance(style_manager, dict):
            instance._style_manager = StyleManager(
                {
                    key: Style.from_ansi(value) 
                    if isinstance(value, str) else value 
                    for key, value in style_manager.items()
                }
            )
        else:
            instance._style_manager = StyleManager()
        instance._styled_text = cls._render(instance)
        return instance

    @property
    def style_manager(self) -> StyleManager:
        """Returns the StyleManager instance associated with this ANSIString."""
        return self._style_manager

    @property
    def styled_text(self) -> str:
        """Returns the styled text, applying styles if they have been modified."""
        if self._style_manager.has_been_modified:
            self._styled_text = self._render()
        return self._styled_text

    @property
    def plain_text(self) -> str:
        """Returns the plain text without any styles."""
        return str.__str__(self)

    @property
    def actual_length(self) -> int:
        """Returns the length of the styled text."""
        return len(self.styled_text)

    def __str__(self) -> str:
        """Returns the styled text."""
        return self.styled_text

    def __repr__(self) -> str:
        """Returns a string representation of the ANSIString."""
        return f"ANSIString({str.__repr__(self.plain_text)}, {self.style_manager if self.style_manager else None})"

    def __eq__(self, other: object) -> bool:
        """Checks if the styled text is equal to another string or ANSIString."""
        return self.styled_text == other

    def __add__(self, other: Union[str, "ANSIString"]) -> "ANSIString":
        """Concatenates another string or ANSIString to this ANSIString."""
        style_manager = self.style_manager.copy()
        if isinstance(other, ANSIString):
            style_manager.update({len(self) + index: value for index, value in other.style_manager.items()})
            other = other.plain_text
        return type(self)(self.plain_text + other, style_manager)

    def __radd__(self, other: Union[str, "ANSIString"]) -> "ANSIString":
        """Concatenates this ANSIString to another string or ANSIString."""
        styles = {index + len(other): value for index, value in self.style_manager.items()}
        if isinstance(other, ANSIString):
            styles.update(other.style_manager)
            other = other.plain_text
        return type(self)(other + self.plain_text, styles)

    def __getitem__(self, key: SupportsIndex | slice) -> "ANSIString":
        """Returns a new ANSIString with the specified slice or index."""
        indices = range(len(self))
        selected_indices = indices[key] if isinstance(key, slice) else [indices[key]]
        styles = {
            new_index: self.style_manager[old_index]
            for new_index, old_index in enumerate(selected_indices)
            if old_index in self.style_manager
        }
        return type(self)(super().__getitem__(key), styles)

    def __getattribute__(self, name: str) -> Any:
        """Handles attribute access, allowing for string methods to return ANSIString."""
        # TODO: Replace it with a more elegant solution like explicit overrides or dynamic class decorator.
        allowed_passthrough = {
            "ljust",
            "rjust",
            "center",
            "split",
            "rsplit",
            "join",
            "splitlines",
        }
        if name in dir(str) and name not in allowed_passthrough:
            def method(self: Self, *args: Any, **kwargs: Any) -> Any:
                result = getattr(str, name)(self.plain_text, *args, **kwargs)

                if isinstance(result, str):
                    return type(self)(result, self.style_manager)
                elif isinstance(result, list):
                    return [type(self)(item, self.style_manager) for item in result]  # type: ignore
                elif isinstance(result, tuple):
                    return tuple(type(self)(item, self.style_manager) for item in result)  # type: ignore
                return result

            return method.__get__(self)
        else:
            return super().__getattribute__(name)

    def __format__(self, format_spec: str) -> str:
        """Formats the ANSIString according to the given format specification and returns rendered text."""
        if not format_spec:
            return self.styled_text
        formatted = format(self.plain_text, format_spec)
        styles = self.style_manager.remap_styles(self.plain_text, formatted)
        return str(type(self)(formatted, StyleManager(styles)))

    def _render(self) -> str:
        """Renders the ANSIString to its final output form."""
        return "".join(
            f"{self.style_manager[index].ansi}{char}\x1b[0m" if index in self.style_manager else char
            for index, char in enumerate(self.plain_text)
        )

    def _coord_to_slice(self, coord: tuple[int, int]) -> slice:
        """Converts a (x, y) coordinate pair to a slice object."""
        index = 0
        lengths = tuple(len(line) for line in self.plain_text.splitlines())
        if not lengths:
            raise IndexError(f"wrong y coordinate (empty string)")
        elif not (0 <= coord[1] < len(lengths)):
            raise IndexError(f"wrong y coordinate (0<=y<{len(lengths)})")
        for y, length in enumerate(lengths):
            if y == coord[1]:
                if not length:
                    raise IndexError(f"wrong x coordinate ({y=}: empty line)")
                elif not (0 <= coord[0] < length):
                    raise IndexError(f"wrong x coordinate ({y=}: 0<=x<{length})")
                index += coord[0] + y
                break
            index += length
        return slice(index, index + 1)

    def _get_all_coords(self) -> tuple[tuple[int, int], ...]:
        """Returns all (x, y) coordinates of the characters in the plain text."""
        def transform(lengths: Iterable[int]) -> Generator[tuple[int, int], None, None]:
            for y, length in enumerate(lengths):
                for x in range(length):
                    yield (x, y)
        return tuple(transform(len(line) for line in self.plain_text.splitlines()))

    def _get_indices(
        self, slice_: Annotated[Sequence[int], Length(3)] | slice
    ) -> tuple[int, int, int]:
        """Converts a slice or a sequence of three integers to start, stop, step indices."""
        if isinstance(slice_, slice):
            start, stop, step = slice_.indices(len(self))
        else:
            start, stop, step = slice(*slice_).indices(len(self))
        return start, stop, step

    def _search_spans(
        self, *words: str, case_sensitive: bool = True
    ) -> tuple[tuple[int, int], ...]:
        """Searches for words in the plain text and returns their spans as tuples of (start, end)."""
        flags = 0 if case_sensitive else re.IGNORECASE
        joined_words = "|".join(re.escape(word) for word in words)
        spans = (match.span(0) for match in re.finditer(joined_words, self.plain_text, flags=flags))
        return tuple(spans)

    def _process_multicolor_command(
        self,
        command: MulticolorCommand,
        rgb: dict[str, dict[str, dict[str, int]]],
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ):
        """Processes a multicolor command and applies it to the ANSIString."""
        if command.reset:
            if command.reset == "?":
                reset_rgb = deepcopy(rgb["actual"])
            elif command.reset == "??":
                reset_rgb = deepcopy(rgb["start"])
        else:
            reset_rgb = None

        modes = {"fg": False, "bg": False, "ul": False}
        for instruction in command.instructions:
            if not modes[instruction.mode]:
                modes[instruction.mode] = True
            if instruction.operator == "=":
                rgb["actual"][instruction.mode][instruction.color] = instruction.processed_value
            elif instruction.operator == "+":
                rgb["actual"][instruction.mode][instruction.color] += instruction.processed_value
            elif instruction.operator == "-":
                rgb["actual"][instruction.mode][instruction.color] -= instruction.processed_value

            rgb["actual"][instruction.mode][instruction.color] = clamp(
                rgb["actual"][instruction.mode][instruction.color], *instruction.minmax
            )

        if slices:
            self._apply_multicolor_command(rgb["actual"], modes, *slices)

        if reset_rgb:
            rgb["actual"] = reset_rgb

        return modes

    def _apply_multicolor_command(
        self,
        rgb: dict[str, dict[str, int]],
        modes: dict[str, bool],
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ) -> None:
        """Applies the current RGB values to the specified slices."""
        if modes["fg"]:
            self.fg_24b(*(int(clamp(rgb["fg"][key], 0, 255)) for key in "rgb"), *slices)
        if modes["bg"]:
            self.bg_24b(*(int(clamp(rgb["bg"][key], 0, 255)) for key in "rgb"), *slices)
        if modes["ul"]:
            self.ul_24b(*(int(clamp(rgb["ul"][key], 0, 255)) for key in "rgb"), *slices)

    @staticmethod
    def from_ansi(plain: str) -> "ANSIString":
        """Creates an ANSIString from a plain string containing ANSI escape sequences."""
        start: int = 0
        decrement: int = 0
        style: str = ""
        styles: dict[int, str] = {}
        sequences: dict[int, str] = {}

        def smart_replacement(match_: re.Match[str]) -> str:
            nonlocal decrement
            sequence, span = match_.group(0), match_.span(0)
            if sequence.endswith("m"):
                if span[0] - decrement in sequences:
                    sequences[span[0] - decrement] += sequence
                else:
                    sequences[span[0] - decrement] = sequence
            decrement += len(sequence)
            return ""

        plain = re.sub(Regex.ANSI_SEQ, smart_replacement, plain)
        for index, sequence in sequences.items():
            for match_ in re.finditer(Regex.SGR_PARAM, sequence):
                parameter = match_.group(0)
                if parameter == "0":
                    if style:
                        for sub_index in range(start, index):
                            styles[sub_index] = style
                        style = ""
                else:
                    style += f"\x1b[{parameter}m"
                    start = index
        return ANSIString(plain, styles)

    def fm(
        self, parameter: int | str, *slices: Annotated[Sequence[int], Length(3)] | slice
    ) -> Self:
        """Formats (applies styling to) the string in a specified range."""
        # TODO: forbid formatting above the length of the string
        if parameter == SGR.RESET:
            return self.unfm(*slices)
        style = Style().with_style(parameter)
        if slices:
            for slice_ in slices:
                for index in range(*self._get_indices(slice_)):
                    if index not in self.style_manager:
                        self.style_manager[index] = style
                    else:
                        self.style_manager[index] = self.style_manager[index].merge(style)
        else:
            for index in range(0, len(self), 1):
                if index not in self.style_manager:
                    self.style_manager[index] = style
                else:
                    self.style_manager[index] = self.style_manager[index].merge(style)
        return self

    def fm_w(
        self, parameter: int | str, *words: str, case_sensitive: bool = True
    ) -> Self:
        """Formats (applies styling to) the word of the string."""
        return self.fm(
            parameter, *self._search_spans(*words, case_sensitive=case_sensitive)
        )

    def unfm(self, *slices: Annotated[Sequence[int], Length(3)] | slice) -> Self:
        """Unformats (removes styling) the string in a specified range."""
        if slices:
            for slice_ in slices:
                for index in range(*self._get_indices(slice_)):
                    if index in self.style_manager:
                        del self.style_manager[index]
        else:
            for index in range(0, len(self), 1):
                if index in self.style_manager:
                    del self.style_manager[index]
        return self

    def unfm_w(self, *words: str, case_sensitive: bool = True) -> Self:
        """Unformats (removes styling) the string per word index."""
        return self.unfm(*self._search_spans(*words, case_sensitive=case_sensitive))

    def fg_4b(
        self,
        parameter: Foreground,
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ) -> Self:
        """Applies the foreground given by the 4-bit color to the string in a specified range."""
        return self.fm(parameter, *slices)

    def fg_4b_w(
        self,
        parameter: Foreground,
        *words: str,
        case_sensitive: bool = True,
    ) -> Self:
        """Applies the foreground given by the 4-bit color to the word of the string."""
        return self.fg_4b(
            parameter, *self._search_spans(*words, case_sensitive=case_sensitive)
        )

    def fg_8b(
        self,
        parameter: Annotated[int, ValueRange(0, 255)],
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ) -> Self:
        """Applies the foreground given by the 8-bit color number (0-255) to the string in a specified range."""
        style = f"\x1b[{Foreground.SET};5;{parameter}m"
        return self.fm(style, *slices)

    def fg_8b_w(
        self,
        parameter: Annotated[int, ValueRange(0, 255)],
        *words: str,
        case_sensitive: bool = True,
    ) -> Self:
        """Applies the foreground given by the 8-bit color number (0-255) to the word of the string."""
        return self.fg_8b(
            parameter, *self._search_spans(*words, case_sensitive=case_sensitive)
        )

    def fg_24b(
        self,
        r: Annotated[int, ValueRange(0, 255)],
        g: Annotated[int, ValueRange(0, 255)],
        b: Annotated[int, ValueRange(0, 255)],
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ) -> Self:
        """Applies the foreground given by RGB to the string in a specified range."""
        style = f"\x1b[{Foreground.SET};2;{r};{g};{b}m"
        return self.fm(style, *slices)

    def fg_24b_w(
        self,
        r: Annotated[int, ValueRange(0, 255)],
        g: Annotated[int, ValueRange(0, 255)],
        b: Annotated[int, ValueRange(0, 255)],
        *words: str,
        case_sensitive: bool = True,
    ) -> Self:
        """Applies the foreground given by RGB to the word of the string."""
        return self.fg_24b(
            r, g, b, *self._search_spans(*words, case_sensitive=case_sensitive)
        )

    def bg_4b(
        self,
        parameter: Background,
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ) -> Self:
        """Applies the background given by the 4-bit color to the string in a specified range."""
        return self.fm(parameter, *slices)

    def bg_4b_w(
        self,
        parameter: Background,
        *words: str,
        case_sensitive: bool = True,
    ) -> Self:
        """Applies the background given by the 4-bit color to the word of the string."""
        return self.bg_4b(
            parameter, *self._search_spans(*words, case_sensitive=case_sensitive)
        )

    def bg_8b(
        self,
        parameter: Annotated[int, ValueRange(0, 255)],
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ) -> Self:
        """Applies the background given by the 8-bit color number (0-255) to the string in a specified range."""
        style = f"\x1b[{Background.SET};5;{parameter}m"
        return self.fm(style, *slices)

    def bg_8b_w(
        self,
        parameter: Annotated[int, ValueRange(0, 255)],
        *words: str,
        case_sensitive: bool = True,
    ) -> Self:
        """Applies the background given by the 8-bit color number (0-255) to the word of the string."""
        return self.bg_8b(
            parameter, *self._search_spans(*words, case_sensitive=case_sensitive)
        )

    def bg_24b(
        self,
        r: Annotated[int, ValueRange(0, 255)],
        g: Annotated[int, ValueRange(0, 255)],
        b: Annotated[int, ValueRange(0, 255)],
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ) -> Self:
        """Applies the background given by RGB to the string in a specified range."""
        style = f"\x1b[{Background.SET};2;{r};{g};{b}m"
        return self.fm(style, *slices)

    def bg_24b_w(
        self,
        r: Annotated[int, ValueRange(0, 255)],
        g: Annotated[int, ValueRange(0, 255)],
        b: Annotated[int, ValueRange(0, 255)],
        *words: str,
        case_sensitive: bool = True,
    ) -> Self:
        """Applies the background given by RGB to the word of the string."""
        return self.bg_24b(
            r, g, b, *self._search_spans(*words, case_sensitive=case_sensitive)
        )
    
    def ul_4b(
        self,
        parameter: Underline,
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ) -> Self:
        """Applies the underline given by the 4-bit color to the string in a specified range."""
        return self.fm(parameter, *slices)

    def ul_4b_w(
        self,
        parameter: Underline,
        *words: str,
        case_sensitive: bool = True,
    ) -> Self:
        """Applies the underline given by the 4-bit color to the word of the string."""
        return self.ul_4b(
            parameter, *self._search_spans(*words, case_sensitive=case_sensitive)
        )

    def ul_8b(
        self,
        parameter: Annotated[int, ValueRange(0, 255)],
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ) -> Self:
        """Applies the underline given by the 8-bit color number (0-255) to the string in a specified range."""
        style = f"\x1b[{Underline.SET}:5:{parameter}m"
        return self.fm(style, *slices)

    def ul_8b_w(
        self,
        parameter: Annotated[int, ValueRange(0, 255)],
        *words: str,
        case_sensitive: bool = True,
    ) -> Self:
        """Applies the underline given by the 8-bit color number (0-255) to the word of the string."""
        return self.ul_8b(
            parameter, *self._search_spans(*words, case_sensitive=case_sensitive)
        )

    def ul_24b(
        self,
        r: Annotated[int, ValueRange(0, 255)],
        g: Annotated[int, ValueRange(0, 255)],
        b: Annotated[int, ValueRange(0, 255)],
        *slices: Annotated[Sequence[int], Length(3)] | slice,
    ) -> Self:
        """Applies the underline given by RGB to the string in a specified range."""
        style = f"\x1b[{Underline.SET}:2::{r}:{g}:{b}m"
        return self.fm(style, *slices)

    def ul_24b_w(
        self,
        r: Annotated[int, ValueRange(0, 255)],
        g: Annotated[int, ValueRange(0, 255)],
        b: Annotated[int, ValueRange(0, 255)],
        *words: str,
        case_sensitive: bool = True,
    ) -> Self:
        """Applies the underline given by RGB to the word of the string."""
        return self.ul_24b(
            r, g, b, *self._search_spans(*words, case_sensitive=case_sensitive)
        )

    def rainbow(
        self,
        *slices: Annotated[Sequence[int], Length(3)] | slice,
        skip_whitespace: bool = False,
        fg: bool = False,
        bg: bool = False,
        ul: bool = False,
    ) -> Self:
        """Applies a rainbow effect to the string in a specified range."""
        if not slices:
            slices = tuple(
                (index, index + 1)
                for index, char in enumerate(self.plain_text)
                if not (skip_whitespace and char in WHITESPACE)
            )
        if not (fg or bg or ul):
            fg = True
        length = len(slices)
        for index, slice_ in enumerate(slices):
            hue = round(index / length * 360)
            if fg:
                self.fg_24b(*hsl_to_rgb(hue), slice_)
            if bg:
                self.bg_24b(*hsl_to_rgb(hue), slice_)
            if ul:
                self.ul_24b(*hsl_to_rgb(hue), slice_)
        return self

    def multicolor(
        self, sequence: str, *slices: Annotated[Sequence[int], Length(3)] | slice
    ) -> Self:
        """Applies a multicolor sequence to the string in a specified range."""
        if not slices:
            slices = tuple((index, index + 1) for index in range(0, len(self)))

        flags = {flag: 0 for flag in ("skipfirst", "cycle", "reverse", "mirror")}
        char_to_flag = {
            "*": "skipfirst",
            "&": "cycle",
            "@": "reverse",
            "!": "mirror",
        }
        offset = 0
        for char in reversed(sequence):
            if char in char_to_flag:
                flags[char_to_flag[char]] = 1
                offset -= 1
            elif char == " ":
                offset -= 1
            else:
                break
        if offset:
            sequence = sequence[:offset]

        rgb = {
            key: {
                key: {key: 0 for key in ("r", "g", "b")} for key in ("fg", "bg", "ul")
            }
            for key in ("actual", "start")
        }

        if "$" in sequence:
            start_command, sequence = map(str.strip, sequence.split("$"))
            object_start_command = MulticolorCommand()
            for start_instruction in map(str.strip, start_command.split("|")):
                match_start_instruction = re.match(
                    Regex.MULTICOLOR_INSTRUCTION, start_instruction
                )
                object_start_instruction = MulticolorInstruction(
                    rgb["actual"],
                    **match_start_instruction.groupdict(),
                    repeat=object_start_command.repeat,
                )
                if match_start_instruction:
                    object_start_command.instructions.append(object_start_instruction)
            start_modes = self._process_multicolor_command(object_start_command, rgb)
            rgb["start"] = deepcopy(rgb["actual"])

        slices_length = len(slices) - (1 if flags["skipfirst"] else 0)
        auto_length = slices_length
        auto_count = 0
        span_decrement = 0
        list_repeats = []
        for match_repeat in re.finditer(r"repeat\((?P<value>\d+|auto)\)", sequence):
            start, stop = match_repeat.span()
            if match_repeat["value"] == "auto":
                auto_count += 1
                list_repeats.append("auto")
            else:
                value = int(repeat["value"])
                if value < auto_length:
                    value = auto_length
                auto_length -= value
                list_repeats.append(value)
            start -= span_decrement
            stop -= span_decrement
            sequence = f"{sequence[:start]}{'{}'}{sequence[stop:]}"
            span_decrement += len(match_repeat.group()) - len(r"{}")
        for index, repeat in enumerate(list_repeats):
            if repeat == "auto":
                value = -(auto_length // -auto_count)
                auto_length -= value
                auto_count -= 1
                list_repeats[index] = value
        sequence = sequence.format(*(f"repeat({value})" for value in list_repeats))

        commands: list[MulticolorCommand] = []
        for command in map(str.strip, sequence.split("#")):
            match_command = re.search(Regex.MULTICOLOR_COMMAND, command)
            object_command = MulticolorCommand(None, **match_command.groupdict())
            if object_command.repeat == 0:
                continue
            for instruction in map(str.strip, command.split("|")):
                match_instruction = re.match(Regex.MULTICOLOR_INSTRUCTION, instruction)
                if match_instruction:
                    object_instruction = MulticolorInstruction(
                        rgb["actual"],
                        **match_instruction.groupdict(),
                        repeat=object_command.repeat,
                    )
                    object_command.instructions.append(object_instruction)
            for no in range(object_command.repeat):
                commands.append(deepcopy(object_command))
                self._process_multicolor_command(object_command, rgb)

        rgb["actual"] = deepcopy(rgb["start"])

        if flags["mirror"] and len(commands) > 1:
            mirrored_commands: list[MulticolorCommand] = []
            for command in reversed(commands):
                mirrored_commands.append(
                    MulticolorCommand(None, command.reset, command.repeat)
                )
                for instruction in command.instructions:
                    copied_instruction = copy(instruction)
                    if copied_instruction.operator == "+":
                        copied_instruction.operator = "-"
                    elif copied_instruction.operator == "-":
                        copied_instruction.operator = "+"
                    mirrored_commands[-1].instructions.append(copied_instruction)
            commands.extend(mirrored_commands)
        elif flags["reverse"]:
            if flags["cycle"] and len(commands) < slices_length:
                for command in cycle(commands):
                    if len(commands) == slices_length:
                        break
                    copied_command = deepcopy(command)
                    for instruction in copied_command.instructions:
                        instruction.process_value(instruction.value, save=True)
                    commands.append(copied_command)
            for command in commands:
                self._process_multicolor_command(command, rgb)
                for instruction in command.instructions:
                    if instruction.operator == "+":
                        instruction.operator = "-"
                    elif instruction.operator == "-":
                        instruction.operator = "+"
            commands.reverse()
        if flags["cycle"] and not flags["reverse"]:
            if len(commands) < slices_length:
                for command in cycle(commands):
                    if len(commands) == slices_length:
                        break
                    copied_command = deepcopy(command)
                    for instruction in copied_command.instructions:
                        instruction.process_value(instruction.value, save=True)
                    commands.append(copied_command)

        if flags["skipfirst"]:
            if (
                slices[0]
                and isinstance(slices[0], Sequence)
                and isinstance(slices[0][0], (Sequence, slice))
            ):
                self._apply_multicolor_command(rgb["actual"], start_modes, *slices[0])
            else:
                self._apply_multicolor_command(rgb["actual"], start_modes, slices[0])
            slices = slices[1:]

        for obj, command in zip(slices, commands):
            if (
                obj
                and isinstance(obj, Sequence)
                and isinstance(obj[0], (Sequence, slice))
            ):
                self._process_multicolor_command(command, rgb, *obj)
            else:
                self._process_multicolor_command(command, rgb, obj)

        return self

    def multicolor_c(self, sequence: str, *coordinates: tuple[int, int]) -> Self:
        """Applies a multicolor sequence to the string at specified (x, y) coordinates."""
        def transform(coordinates):
            for obj in coordinates:
                if isinstance(obj, tuple) and isinstance(obj[0], tuple):
                    yield tuple(self._coord_to_slice(coord) for coord in obj)
                else:
                    yield self._coord_to_slice(obj)

        if not coordinates:
            coordinates = self._get_all_coords()
        return self.multicolor(sequence, *transform(coordinates))

    def to_svg(
        self,
        font: TTFont | Path | str,
        font_size_px: int | float,
        line_height_offset: int | float = 0,
        letter_spacing_offset: int | float = 0,
        transparent_background: bool = True,
        background_color: tuple[int, int, int] = (255, 255, 255),
        convert_text_to_path: bool = False,
        output_file: str | None = None,
    ) -> str:
        """Generates an SVG representation of the ANSIString using the specified font."""
        if not is_fonttools_available:
            raise ImportError(
                "The 'fontTools' package is required to use the 'to_svg' method. "
                "Please install it using 'pip install fonttools'."
            )
        if isinstance(font, (Path, str)):
            font = TTFont(font)

        escape_table = {
            "&": "&amp;",
            "<": "&lt;",
            ">": "&gt;",
            '"': "&quot;",
            "'": "&#39;",
        }
        underline_style_map = {
            UnderlineMode.SINGLE: "solid",
            UnderlineMode.DOUBLE: "double",
            UnderlineMode.DOTTED: "dotted",
            UnderlineMode.DASHED: "dashed",
            UnderlineMode.CURLY: "wavy",
        }
        lines = self.plain_text.splitlines(keepends=False)
        svg_parts: list[str] = []
        texts: list[str] = []
        paths: list[str] = []
        rects: list[str] = []
        chars: list[str] = []

        font_family = font['name'].getDebugName(1) or "sans-serif"

        cmap = font.getBestCmap()
        glyph_set = font.getGlyphSet()

        units_per_em = font['head'].unitsPerEm
        ascent = font['hhea'].ascent
        descent = font['hhea'].descent
        line_gap = font['hhea'].lineGap

        hmtx = font["hmtx"]

        scale = font_size_px / units_per_em

        em_height = ascent - descent
        line_height = em_height + line_gap + line_height_offset
        line_height_px = line_height * scale

        total_width = 0
        total_height = line_height * len(lines) * scale

        charno = 0
        y_cursor = ascent + line_gap / 2 + line_height_offset / 2
        for lineno, line in enumerate(lines):
            x_cursor = 0
            for char in line:
                escaped_char: str = escape_table.get(char, char)

                glyph_name = cmap.get(ord(char), ".notdef")
                glyph = glyph_set.get(glyph_name, glyph_set[".notdef"])
                advance_width, _ = hmtx[glyph_name]

                if charno in self.style_manager:
                    svg_element_attributes: list[str] = []

                    if self.style_manager[charno].background:
                        rects.append(
                            " "*2
                            + f"<rect "
                            + f"x=\"{x_cursor * scale}\" "
                            + f"y=\"{(y_cursor - ascent - line_gap / 2 - line_height_offset / 2) * scale}\" "
                            + f"width=\"{(advance_width + letter_spacing_offset) * scale}\" "
                            + f"height=\"{line_height_px}\" "
                            + f"fill=\"rgb{self.style_manager[charno].background.to_rgb()}\""
                            + "/>"
                        )
                    if self.style_manager[charno].foreground:
                        svg_element_attributes.append(
                            f"fill=\"rgb{self.style_manager[charno].foreground.to_rgb()}\""
                        )

                    if not convert_text_to_path:
                        if SGR.BOLD in self.style_manager[charno].attributes:
                            svg_element_attributes.append(
                                "font-weight=\"bold\""
                            )
                        if SGR.ITALIC in self.style_manager[charno].attributes:
                            svg_element_attributes.append(
                                "font-style=\"italic\""
                            )
                        if not self.style_manager[charno].underline[0]:
                            if SGR.UNDERLINE in self.style_manager[charno].attributes:
                                svg_element_attributes.append(
                                    "text-decoration=\"underline auto solid\""
                                )
                            elif SGR.DOUBLE_UNDERLINE in self.style_manager[charno].attributes:
                                svg_element_attributes.append(
                                    "text-decoration=\"underline auto double\""
                                )
                            chars.append(
                                f"<tspan{(' ' + ' '.join(svg_element_attributes)) if svg_element_attributes else ''}>"
                                + escaped_char
                                + "</tspan>"
                            )
                        else:
                            if not self.style_manager[charno].foreground:
                                svg_element_attributes.append(
                                    f"fill=\"currentColor\""
                                )
                            chars.append(
                                f"<tspan "
                                + f"fill=\"rgb{self.style_manager[charno].underline[0].to_rgb()}\" "
                                + f"text-decoration=\"underline auto {underline_style_map.get(self.style_manager[charno].underline[1], 'solid')}\""
                                + ">"
                                + f"<tspan{(' ' + ' '.join(svg_element_attributes)) if svg_element_attributes else ''}>"
                                + escaped_char
                                + ("</tspan>"*2)
                            )
                    else:
                        pen = SVGPathPen(glyph_set)
                        t_pen = TransformPen(pen, (scale, 0, 0, -scale, x_cursor * scale, y_cursor * scale))
                        
                        glyph_set[glyph_name].draw(t_pen)
                        svg_element_attributes.append(
                            f"d=\"{pen.getCommands()}\""
                        ) # FIXME: check for whitespace

                        ... # TODO: bold, italic, underline solution for path

                        paths.append(
                            f"{' '*2}<path {' '.join(svg_element_attributes)}/>"
                        )
                else:
                    if not convert_text_to_path:
                        chars.append(f"<tspan>{escape_table.get(char, char)}</tspan>")
                    else:
                        pen = SVGPathPen(glyph_set)
                        t_pen = TransformPen(pen, (scale, 0, 0, -scale, x_cursor * scale, y_cursor * scale))
                        
                        glyph_set[glyph_name].draw(t_pen)
                        
                        paths.append(
                            f"{' '*2}<path d=\"{pen.getCommands()}\"/>"
                        )

                x_cursor += advance_width + letter_spacing_offset
                charno += 1  # character
            
            if not convert_text_to_path:
                texts.append(
                    f"<tspan x=\"0\" dy=\"{line_height_offset / 2 * scale if lineno == 0 else line_height_px}\">{''.join(chars)}</tspan>"
                )
                chars.clear()

            if x_cursor * scale > total_width:
                total_width = x_cursor * scale
            y_cursor += line_height
            charno += 1  # newline

        svg_parts = [
            # "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>",
            f'<svg xmlns="http://www.w3.org/2000/svg" width="{total_width}" height="{total_height}" viewBox="0 0 {total_width} {total_height}">'
            + (f'\n{" "*2}<rect width="100%" height="100%" fill="rgb{background_color}"/>' if not transparent_background else "")
        ]
        for svg_rect in rects:
            svg_parts.append(svg_rect)
        if not convert_text_to_path:
            svg_parts.append(
                f'{" "*2}<text x=\"{0}\" y=\"{(ascent + line_gap / 2) * scale}\" font-family=\"{font_family}\" font-size=\"{font_size_px}\" fill=\"black\" letter-spacing=\"{letter_spacing_offset * scale}\">'
                + "".join(texts)
                + f"</text>"
            )
        for svg_path in paths:
            svg_parts.append(svg_path)
        svg_parts.append("</svg>")

        svg_content ="\n".join(svg_parts)

        if output_file:
            with open(output_file, "wt", encoding="utf-8") as file:
                file.write(svg_content)

        return svg_content

    def join(self, iterable: Iterable[str], /) -> "ANSIString":
        styles: dict[int, Style] = {}
        increment = 0
        for i, string in enumerate(iterable):
            increment += len(string)
            if i:
                increment += len(self)
            styles.update(
                {increment + index: style for index, style in self.style_manager.items()}
            )
            if type(string) == ANSIString:
                styles.update(
                    {
                        increment + index - len(string): style
                        for index, style in string.style_manager.items()
                    }
                )
        return type(self)(super().join(iterable), StyleManager(styles))

    def ljust(self, width: SupportsIndex, fillchar: str = " ") -> "ANSIString":
        return self + fillchar * (int(width) - len(self))

    def rjust(self, width: SupportsIndex, fillchar: str = " ") -> "ANSIString":
        return fillchar * (int(width) - len(self)) + self  # type: ignore

    def center(self, width: SupportsIndex, fillchar: str = " ") -> "ANSIString":
        margin = int(width) - len(self)
        left = (margin // 2) + (margin & int(width) & 1)
        return fillchar * left + self + fillchar * (margin - left)  # type: ignore

    def rsplit(self, sep: str | None = None, maxsplit: SupportsIndex = -1) -> list["ANSIString"]:  # type: ignore
        actual = super().rsplit(sep, maxsplit)
        max_index = len(self)
        if not sep:
            whitespace = rsearch_separators(self.plain_text)
            if self.plain_text[-1] in WHITESPACE:
                max_index -= len(next(whitespace, ""))
        for no, string in enumerate(actual[::-1]):
            min_index = max_index - len(string)
            styles = {
                index - min_index: self.style_manager[index]
                for index in range(min_index, max_index)
                if index in self.style_manager
            }
            actual[len(actual) - 1 - no] = type(self)(string, StyleManager(styles))
            max_index -= len(string) + (len(sep) if sep else len(next(whitespace, "")))  # type: ignore
        return actual  # type: ignore

    def split(self, sep: str | None = None, maxsplit: SupportsIndex = -1) -> list["ANSIString"]:  # type: ignore
        actual = super().split(sep, maxsplit)
        min_index = 0
        if not sep:
            whitespace = search_separators(self.plain_text)
            if self.plain_text[0] in WHITESPACE:
                min_index += len(next(whitespace, ""))
        for no, string in enumerate(actual):
            max_index = min_index + len(string)
            styles = {
                index - min_index: self.style_manager[index]
                for index in range(min_index, max_index)
                if index in self.style_manager
            }
            actual[no] = type(self)(string, StyleManager(styles))
            min_index += len(string) + (len(sep) if sep else len(next(whitespace, "")))  # type: ignore
        return actual  # type: ignore

    def splitlines(self, keepends: bool = False) -> list["ANSIString"]:  # type: ignore
        actual = super().splitlines(keepends)
        min_index = 0
        for no, string in enumerate(actual):
            max_index = min_index + len(string)
            styles = {
                index - min_index: self.style_manager[index]
                for index in range(min_index, max_index)
                if index in self.style_manager
            }
            actual[no] = type(self)(string, StyleManager(styles))
            min_index += len(string) + (0 if keepends else 1)
        return actual  # type: ignore



from functools import wraps
from types import MethodType
from typing import Any, Callable, TypeVar, ParamSpec, cast

from .style import Style


P = ParamSpec("P")
R = TypeVar("R")

def _detect_style_change(
    method: Callable[P, R],
    is_method_bound: bool = False
) -> Callable[P, R]:
    """Decorator to detect changes in the StyleManager and set the modified flag."""
    @wraps(method)
    def wrapped(self: "StyleManager", *args: P.args, **kwargs: P.kwargs) -> R:
        previous_length = len(self)
        if not is_method_bound:
            result = method(self, *args, **kwargs)  # type: ignore
        else:
            result = method(*args, **kwargs)
        if not self._has_been_modified and previous_length != len(self):  # type: ignore
            self._has_been_modified = True  # type: ignore
        return result # type: ignore
    return cast(Callable[P, R], wrapped)

class StyleManager(dict[int, Style]):
    """
    A dictionary-like class for managing `Style` instances and tracking style modifications.

    This class behaves similarly to a dictionary, allowing storage and retrieval of `Style`
    instances. It also maintains an internal flag to indicate whether any style has been modified 
    since the last access.

    Attributes
    ----------
    _has_been_modified : bool
        Indicates whether the styles have been modified since the last check.
    has_been_modified : bool
        Property. Returns `True` if styles have been modified since the last access;
        accessing this resets the flag to `False`.

    Examples
    --------
    >>> style_manager = StyleManager()
    >>> style_manager[key] = value
    >>> style_manager.has_been_modified
    True
    >>> style_manager.has_been_modified
    False
    """
    _style_cache: dict[int, Style] = {}

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._has_been_modified = False
        # TODO: wrap methods manually, with @is_method_bound
        for name in {"clear", "pop", "popitem", "setdefault", "update"}:
            method = getattr(self, name)
            wrapped = _detect_style_change(method, is_method_bound=True)
            setattr(self, name, MethodType(wrapped, self))  # NOTE: wrapped.__get__(self, self.__class__)

    @property
    def has_been_modified(self) -> bool:
        """Check if the styles have been modified since the last access."""
        result = self._has_been_modified
        if self._has_been_modified:
            self._has_been_modified = False
        return result

    def __repr__(self) -> str:
        """Return a string representation of the StyleManager."""
        # TODO: it is too verbose, but it is useful for debugging
        return f"StyleManager({dict.__repr__(self)})"  # type: ignore

    def __setitem__(self, key: Any, value: Any) -> None:
        """ Set a `Style` instance in the dictionary."""
        if not isinstance(value, Style):
            raise TypeError("StyleManager values must be Style instances")
        # NOTE: Cache identical Style objects by their hash
        style_hash = hash(value)
        cached = self._style_cache.get(style_hash)
        if cached is not None and cached == value:
            value = cached
        else:
            self._style_cache[style_hash] = value
        self._has_been_modified = True
        return super().__setitem__(key, value)

    @_detect_style_change
    def __delitem__(self, key: Any) -> None:
        """Delete a style from the dictionary."""
        return super().__delitem__(key)

    def copy(self) -> "StyleManager":
        """Create a shallow copy of the StyleManager."""
        copied = StyleManager(dict[Any, Any].copy(self))
        copied._has_been_modified = self._has_been_modified
        return copied

    def remap_styles(
        self,
        original: str,
        formatted: str,
        visible_only: bool = True
    ) -> dict[int, Style]:
        """Remap styles from the original string to the formatted string."""
        if formatted == original:
            return dict(self)

        pad_left = formatted.find(original)
        if pad_left == -1:
            raise ValueError("Original string not found inside formatted string.")

        if visible_only:
            # Copy only styles that fall within the visible range of `original` inside `formatted`
            styles = {
                index: self[index - pad_left]
                for index in range(pad_left, pad_left + len(original))
                if (index - pad_left) in self  # avoid KeyError if self is missing some indexes
            }
        else:
            # Remap all style indexes, shifted by pad_left
            styles = {
                index + pad_left: self[index]
                for index in self.keys()
            }

        return styles
